"""Andrews-Roth-Pakes (APR) confidence intervals with no nuisance parameters."""

import warnings
from typing import NamedTuple

import numpy as np

from .conditional import _norminvp_generalized
from .numba import compute_bounds, prepare_theta_grid_y_values, selection_matrix
from .utils import basis_vector


class APRCIResult(NamedTuple):
    """Result from APR confidence interval computation.

    Attributes
    ----------
    ci_lower : float
        Lower bound of confidence interval.
    ci_upper : float
        Upper bound of confidence interval.
    ci_length : float
        Length of confidence interval.
    theta_grid : ndarray
        Grid of theta values tested.
    accept_grid : ndarray
        Boolean array indicating which theta values are in the identified set.
    status : str
        Optimization status.
    """

    ci_lower: float
    ci_upper: float
    ci_length: float
    theta_grid: np.ndarray
    accept_grid: np.ndarray
    status: str


def compute_arp_ci(
    beta_hat,
    sigma,
    A,
    d,
    n_pre_periods,
    n_post_periods,
    post_period_index=1,
    alpha=0.05,
    grid_lb=None,
    grid_ub=None,
    grid_points=1000,
    return_length=False,
    hybrid_flag="ARP",
    hybrid_kappa=None,
    flci_halflength=None,
    flci_l=None,
    lf_cv=None,
):
    r"""Compute Andrews-Roth-Pakes (ARP) confidence interval with no nuisance parameters.

    Constructs confidence intervals for the parameter of interest :math:`\theta = l' \tau_{\text{post}}`
    in the special case where :math:`\bar{T} = 1` (single post-treatment period), which means
    there are no nuisance parameters :math:`\tilde{\tau}` to profile over. This allows for more
    efficient computation compared to the general case.

    For each value :math:`\bar{\theta}` on a grid, the method tests the null hypothesis
    :math:`H_0: \theta = \bar{\theta}, \delta \in \Delta` where :math:`\Delta = \{\delta: A \delta \leq d\}`.
    Following the results from [1]_, this is equivalent to testing whether there exists
    :math:`\tau_{\text{post}} \in \mathbb{R}^{\bar{T}}` such that :math:`l' \tau_{\text{post}} = \bar{\theta}`
    and

    .. math::

        \mathbb{E}_{\hat{\beta}_n \sim \mathcal{N}(\delta+\tau, \Sigma_n)}
            [Y_n - A L_{\text{post}} \tau_{\text{post}}] \leq 0

    where :math:`Y_n = A \hat{\beta}_n - d` and :math:`L_{\text{post}} = [0, I]'`.

    In the no-nuisance case (:math:`\bar{T} = 1`), the profiled test statistic from equation (14) in [2]_
    simplifies to :math:`\hat{\eta} = \max_i (A \hat{\beta}_n - d)_i / \tilde{\sigma}_{n,i}` where
    :math:`\tilde{\sigma}_{n,i} = \sqrt{(A \Sigma_n A')_{ii}}`. The test conditions on the event
    that constraint :math:`j = \arg\max_i (A \hat{\beta}_n - d)_i / \tilde{\sigma}_{n,i}` is binding,
    leading to a truncated normal distribution for :math:`\hat{\eta}`.

    Parameters
    ----------
    beta_hat : ndarray
        Vector of estimated event study coefficients :math:`\hat{\beta}`. First
        `n_pre_periods` elements are pre-treatment, remainder are post-treatment.
    sigma : ndarray
        Covariance matrix :math:`\Sigma` of estimated coefficients.
    A : ndarray
        Matrix :math:`A` defining the constraint set :math:`\Delta`.
    d : ndarray
        Vector :math:`d` such that :math:`\Delta = \{\delta : A\delta \leq d\}`.
    n_pre_periods : int
        Number of pre-treatment periods :math:`T_{pre}`.
    n_post_periods : int
        Number of post-treatment periods :math:`T_{post}`.
    post_period_index : int, default=1
        Which post-treatment period :math:`s` to compute CI for (1-indexed).
    alpha : float, default=0.05
        Significance level :math:`\alpha` for confidence interval.
    grid_lb : float, optional
        Lower bound for grid search. If None, uses :math:`\hat{\theta} - 10 \cdot SE(\hat{\theta})`.
    grid_ub : float, optional
        Upper bound for grid search. If None, uses :math:`\hat{\theta} + 10 \cdot SE(\hat{\theta})`.
    grid_points : int, default=1000
        Number of points in the grid search.
    return_length : bool, default=False
        If True, only return the CI length (useful for power calculations).
    hybrid_flag : {'ARP', 'FLCI', 'LF'}, default='ARP'
        Type of test to use. 'ARP' is the standard conditional test, 'FLCI' adds
        fixed-length CI constraints for improved power, 'LF' uses a least favorable
        first stage.
    hybrid_kappa : float, optional
        First-stage size :math:`\kappa` for hybrid tests. Required if hybrid_flag != 'ARP'.
    flci_halflength : float, optional
        Half-length of FLCI constraint. Required if hybrid_flag == 'FLCI'.
    flci_l : ndarray, optional
        Weight vector :math:`\ell` for FLCI. Required if hybrid_flag == 'FLCI'.
    lf_cv : float, optional
        Critical value for LF test. Required if hybrid_flag == 'LF'.

    Returns
    -------
    APRCIResult
        NamedTuple containing CI bounds, grid of tested values, acceptance
        indicators, and optimization status.

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2023). Inference for Linear
        Conditional Moment Inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    beta_hat = np.asarray(beta_hat).flatten()
    sigma = np.asarray(sigma)
    A = np.asarray(A)
    d = np.asarray(d).flatten()

    if beta_hat.shape[0] != n_pre_periods + n_post_periods:
        raise ValueError(
            f"beta_hat length ({beta_hat.shape[0]}) must equal "
            f"n_pre_periods + n_post_periods ({n_pre_periods + n_post_periods})"
        )

    if sigma.shape[0] != sigma.shape[1] or sigma.shape[0] != beta_hat.shape[0]:
        raise ValueError("sigma must be square and conformable with beta_hat")

    if post_period_index < 1 or post_period_index > n_post_periods:
        raise ValueError(f"post_period_index must be between 1 and {n_post_periods}")

    if hybrid_flag != "ARP":
        if hybrid_kappa is None:
            raise ValueError(f"hybrid_kappa must be specified for {hybrid_flag} test")
        if hybrid_flag == "FLCI":
            if flci_halflength is None or flci_l is None:
                raise ValueError("flci_halflength and flci_l must be specified for FLCI hybrid")
        elif hybrid_flag == "LF" and lf_cv is None:
            raise ValueError("lf_cv must be specified for LF hybrid")

    if grid_lb is None:
        post_period_vec = basis_vector(index=n_pre_periods + post_period_index, size=len(beta_hat)).flatten()
        point_est = post_period_vec @ beta_hat
        se = np.sqrt(post_period_vec @ sigma @ post_period_vec)
        grid_lb = point_est - 10 * se

    if grid_ub is None:
        post_period_vec = basis_vector(index=n_pre_periods + post_period_index, size=len(beta_hat)).flatten()
        point_est = post_period_vec @ beta_hat
        se = np.sqrt(post_period_vec @ sigma @ post_period_vec)
        grid_ub = point_est + 10 * se

    theta_grid = np.linspace(grid_lb, grid_ub, grid_points)

    if hybrid_flag == "ARP":
        test_fn = test_in_identified_set
    elif hybrid_flag == "FLCI":
        test_fn = test_in_identified_set_flci_hybrid
    elif hybrid_flag == "LF":
        test_fn = test_in_identified_set_lf_hybrid
    else:
        raise ValueError(f"Invalid hybrid_flag: {hybrid_flag}")

    results_grid = _test_over_theta_grid(
        beta_hat=beta_hat,
        sigma=sigma,
        A=A,
        d=d,
        theta_grid=theta_grid,
        n_pre_periods=n_pre_periods,
        post_period_index=post_period_index,
        alpha=alpha,
        test_fn=test_fn,
        hybrid_kappa=hybrid_kappa,
        flci_halflength=flci_halflength,
        flci_l=flci_l,
        lf_cv=lf_cv,
    )

    accept_grid = results_grid[:, 1].astype(bool)
    accepted_thetas = theta_grid[accept_grid]

    if accept_grid[0] or accept_grid[-1]:
        warnings.warn(
            "CI is open at one of the endpoints; CI bounds may not be accurate. Consider expanding the grid bounds.",
            UserWarning,
        )

    if len(accepted_thetas) == 0:
        ci_lower = np.nan
        ci_upper = np.nan
        ci_length = np.nan
        status = "empty_ci"
    else:
        ci_lower = np.min(accepted_thetas)
        ci_upper = np.max(accepted_thetas)
        ci_length = ci_upper - ci_lower
        status = "success"

    if return_length:
        return ci_length

    return APRCIResult(
        ci_lower=ci_lower,
        ci_upper=ci_upper,
        ci_length=ci_length,
        theta_grid=theta_grid,
        accept_grid=accept_grid,
        status=status,
    )


def test_in_identified_set(
    y,
    sigma,
    A,
    d,
    alpha,
    _precomputed=None,
    **kwargs,
):
    r"""Test whether :math:`\bar{\theta}` lies in the identified set using the ARP conditional approach.

    Tests the null hypothesis :math:`H_0: \theta = \bar{\theta}, \delta \in \Delta` by checking
    whether the moment inequalities :math:`\mathbb{E}[\tilde{Y}_n(\bar{\theta}) - \tilde{X}\tilde{\tau}] \leq 0`
    hold for some :math:`\tilde{\tau}` (equation 13 in [2]_). In the no-nuisance case where :math:`\bar{T} = 1`,
    there is no :math:`\tilde{\tau}` to optimize over.

    Following equations (14)-(15) in [2]_, the test statistic is the solution to the dual program

    .. math::
        \hat{\eta} = \max_{\gamma} \gamma' \tilde{Y}_n(\bar{\theta})
        \text{ s.t. } \gamma' \tilde{X} = 0, \gamma' \tilde{\sigma}_n = 1, \gamma \geq 0,

    where :math:`\tilde{Y}_n(\bar{\theta}) = A\hat{\beta}_n - d - A L_{\text{post}}(\bar{\theta}, 0)'`
    and :math:`\tilde{\sigma}_n = \sqrt{\text{diag}(A \Sigma_n A')}`. In the no-nuisance case,
    the constraint :math:`\gamma' \tilde{X} = 0` is vacuous, so :math:`\hat{\eta}` simplifies to
    :math:`\max_i (\tilde{Y}_n(\bar{\theta}))_i / \tilde{\sigma}_{n,i}`.

    The test conditions on the event :math:`\{\gamma_* \in \hat{V}_n, S_n = s\}` where :math:`\gamma_*`
    is the optimal vertex. Under this conditioning, :math:`\hat{\eta}` follows a truncated normal
    distribution with truncation bounds :math:`[v^{lo}, v^{up}]` that ensure :math:`\gamma_*` remains optimal.

    Parameters
    ----------
    y : ndarray
        Observed coefficient vector :math:`Y = \hat{\beta} - \theta_0 e_{post,s}`
        where :math:`\theta_0` is the hypothesized value.
    sigma : ndarray
        Covariance matrix :math:`\Sigma` of the event study coefficients.
    A : ndarray
        Constraint matrix :math:`A` defining :math:`\Delta`.
    d : ndarray
        Constraint bounds :math:`d` such that :math:`\Delta = \{\delta : A\delta \leq d\}`.
    alpha : float
        Significance level :math:`\alpha` for the test.
    _precomputed : dict, optional
        Pre-computed A_tilde and d_tilde to avoid redundant computation in grid loops.
    **kwargs
        Unused parameters for compatibility with hybrid tests.

    Returns
    -------
    bool
        True if null is NOT rejected (i.e., :math:`\theta_0` is in the confidence set).

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2023). Inference for Linear
        Conditional Moment Inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    if _precomputed is not None:
        A_tilde = _precomputed["A_tilde"]
        d_tilde = _precomputed["d_tilde"]
    else:
        sigma_tilde = np.sqrt(np.diag(A @ sigma @ A.T))
        sigma_tilde = np.maximum(sigma_tilde, 1e-10)
        A_tilde = np.diag(1 / sigma_tilde) @ A
        d_tilde = d / sigma_tilde

    normalized_moments = A_tilde @ y - d_tilde
    max_location = np.argmax(normalized_moments)
    max_moment = normalized_moments[max_location]

    # If max_moment is positive, we have a constraint violation
    # In this case, we need to check if it's statistically significant
    if max_moment <= 0:
        return True

    # Construct conditioning event
    T_B = selection_matrix([max_location + 1], size=len(normalized_moments), select="rows")
    iota = np.ones((len(normalized_moments), 1))

    gamma = A_tilde.T @ T_B.T
    A_bar = A_tilde - iota @ T_B @ A_tilde
    d_bar = (np.eye(len(d_tilde)) - iota @ T_B) @ d_tilde

    # Compute conditional distribution parameters
    sigma_bar = np.sqrt(gamma.T @ sigma @ gamma).item()
    c = sigma @ gamma / (gamma.T @ sigma @ gamma).item()
    z = (np.eye(len(y)) - c @ gamma.T) @ y

    v_lo, v_up = compute_bounds(eta=gamma, sigma=sigma, A=A_bar, b=d_bar, z=z)

    # Check if the observed max_moment is within the truncation bounds
    # If max_moment < v_lo, then the observed value is outside the conditional support
    # and we should reject (this point cannot arise under the null)
    if max_moment < v_lo:
        # The observed value is impossible under the null hypothesis
        # given the conditioning event, so we reject
        return False

    critical_val = max(
        0,
        _norminvp_generalized(
            p=1 - alpha,
            lower=v_lo,
            upper=v_up,
            mu=(T_B @ d_tilde).item(),
            sd=sigma_bar,
        ),
    )

    reject = max_moment > critical_val
    return not reject


def test_in_identified_set_flci_hybrid(
    y,
    sigma,
    A,
    d,
    alpha,
    hybrid_kappa,
    flci_halflength,
    flci_l,
    **kwargs,
):
    r"""Hybrid test combining fixed-length confidence interval (FLCI) constraints with ARP conditional test.

    Implements a two-stage hybrid test following the general structure described in Section 3.2
    of [2]_. The first stage checks whether the FLCI constraints :math:`|\ell' Y| \leq h_{FLCI}`
    are satisfied, where :math:`\ell` is an optimally chosen weight vector and :math:`h_{FLCI}`
    is the FLCI half-length. If these constraints are violated (i.e., if :math:`|\ell' Y| > h_{FLCI}`),
    the test rejects immediately with size :math:`\kappa`.

    If the first stage does not reject, the second stage proceeds with a modified conditional
    test that includes the FLCI constraints in the constraint set. Following the hybrid approach,
    the second stage uses adjusted size :math:`\tilde{\alpha} = \frac{\alpha - \kappa}{1 - \kappa}`
    to ensure overall size :math:`\alpha` control.

    The FLCI constraints :math:`|\ell' Y| \leq h_{FLCI}` are reformulated as two linear inequalities:
    :math:`\ell' Y \leq h_{FLCI}` and :math:`-\ell' Y \leq h_{FLCI}`, which are added to the
    original constraint set :math:`\Delta = \{\delta : A\delta \leq d\}` for the second stage test.

    Parameters
    ----------
    y : ndarray
        Observed coefficient vector :math:`Y = \hat{\beta} - \theta_0 e_{post,s}`.
    sigma : ndarray
        Covariance matrix :math:`\Sigma`.
    A : ndarray
        Constraint matrix :math:`A` for main restrictions.
    d : ndarray
        Constraint bounds :math:`d`.
    alpha : float
        Overall significance level :math:`\alpha`.
    hybrid_kappa : float
        First-stage significance level :math:`\kappa`.
    flci_halflength : float
        Half-length :math:`h_{FLCI}` of the fixed-length confidence interval.
    flci_l : ndarray
        Weight vector :math:`\ell` from FLCI optimization.
    **kwargs
        Unused parameters for compatibility.

    Returns
    -------
    bool
        True if null is NOT rejected (value is in identified set).

    Notes
    -----
    The FLCI hybrid leverages the optimal linear combination :math:`\ell` found
    by minimizing worst-case CI length. This often provides tighter bounds than
    the least favorable approach, especially when :math:`\Delta` has special
    structure like smoothness restrictions.

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2023). Inference for Linear
        Conditional Moment Inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    # First stage: test FLCI constraint
    flci_l = np.asarray(flci_l).flatten()

    # Create constraints for |l'y| <= halflength
    A_firststage = np.vstack([flci_l, -flci_l])
    d_firststage = np.array([flci_halflength, flci_halflength])

    if np.max(A_firststage @ y - d_firststage) > 0:
        return False

    # Second stage: run modified APR test
    # Adjust significance level
    alpha_tilde = (alpha - hybrid_kappa) / (1 - hybrid_kappa)

    # Add first-stage constraints to main constraints
    A_combined = np.vstack([A, A_firststage])
    d_combined = np.hstack([d, d_firststage])

    return test_in_identified_set(
        y=y,
        sigma=sigma,
        A=A_combined,
        d=d_combined,
        alpha=alpha_tilde,
    )


def test_in_identified_set_lf_hybrid(
    y,
    sigma,
    A,
    d,
    alpha,
    hybrid_kappa,
    lf_cv,
    _precomputed=None,
    **kwargs,
):
    r"""Conditional-least favorable (LF) hybrid test.

    Implements the conditional-LF hybrid test that combines a least favorable (LF) first stage
    with a conditional second stage. As described in [1]_, the distribution of :math:`\hat{\eta}`
    under the null is bounded above (in the sense of first-order stochastic dominance) by
    the distribution when :math:`\tilde{\mu}(\bar{\theta}) = 0`.

    The first stage uses a size-:math:`\kappa` LF test that rejects when
    :math:`\hat{\eta} > c_{LF,\kappa}`, where :math:`c_{LF,\kappa}` is the :math:`1-\kappa`
    quantile of :math:`\max_{\gamma \in V(\Sigma)} \gamma' \xi` with :math:`\xi \sim \mathcal{N}(0, \tilde{\Sigma}_n)`.
    This critical value can be calculated by simulation as it depends only on :math:`\tilde{\Sigma}_n`.

    If the first stage does not reject, the second stage conducts a modified conditional test
    with size :math:`\frac{\alpha - \kappa}{1 - \kappa}` that also conditions on the event
    :math:`\{\hat{\eta} \leq c_{LF,\kappa}\}`. The truncation upper bound becomes
    :math:`v_H^{up} = \min\{v^{up}, c_{LF,\kappa}\}`, ensuring the test conditions on passing
    the first stage.

    This hybrid approach improves power when binding and non-binding moments are close together
    (relative to sampling variation) while maintaining exact size :math:`\alpha` control.

    Parameters
    ----------
    y : ndarray
        Observed coefficient vector :math:`Y = \hat{\beta} - \theta_0 e_{post,s}`.
    sigma : ndarray
        Covariance matrix :math:`\Sigma`.
    A : ndarray
        Constraint matrix :math:`A`.
    d : ndarray
        Constraint bounds :math:`d`.
    alpha : float
        Overall significance level :math:`\alpha`.
    hybrid_kappa : float
        First-stage significance level :math:`\kappa`, typically :math:`\alpha/10`.
    lf_cv : float
        Least favorable critical value :math:`c_{LF}` for first-stage test.
    _precomputed : dict, optional
        Pre-computed A_tilde and d_tilde to avoid redundant computation in grid loops.
    **kwargs
        Unused parameters.

    Returns
    -------
    bool
        True if null is NOT rejected (value is in identified set).

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2023). Inference for Linear
        Conditional Moment Inequalities. Review of Economic Studies.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to
        parallel trends. Review of Economic Studies, 90(5), 2555-2591.
    """
    if _precomputed is not None:
        A_tilde = _precomputed["A_tilde"]
        d_tilde = _precomputed["d_tilde"]
    else:
        sigma_tilde = np.sqrt(np.diag(A @ sigma @ A.T))
        sigma_tilde = np.maximum(sigma_tilde, 1e-10)
        A_tilde = np.diag(1 / sigma_tilde) @ A
        d_tilde = d / sigma_tilde

    normalized_moments = A_tilde @ y - d_tilde
    max_location = np.argmax(normalized_moments)
    max_moment = normalized_moments[max_location]

    # First stage test
    if max_moment > lf_cv:
        return False

    # Second stage: condition on passing first stage
    # Construct conditioning event as before
    T_B = selection_matrix([max_location + 1], size=len(normalized_moments), select="rows")
    iota = np.ones((len(normalized_moments), 1))

    gamma = A_tilde.T @ T_B.T
    A_bar = A_tilde - iota @ T_B @ A_tilde
    d_bar = (np.eye(len(d_tilde)) - iota @ T_B) @ d_tilde

    sigma_bar = np.sqrt(gamma.T @ sigma @ gamma).item()
    c = sigma @ gamma / (gamma.T @ sigma @ gamma).item()
    z = (np.eye(len(y)) - c @ gamma.T) @ y

    v_lo, v_up = compute_bounds(eta=gamma, sigma=sigma, A=A_bar, b=d_bar, z=z)
    alpha_tilde = (alpha - hybrid_kappa) / (1 - hybrid_kappa)

    critical_val = max(
        0,
        _norminvp_generalized(
            p=1 - alpha_tilde,
            lower=v_lo,
            upper=v_up,
            mu=(T_B @ d_tilde).item(),
            sd=sigma_bar,
        ),
    )

    reject = max_moment > critical_val
    return not reject


def _test_over_theta_grid(
    beta_hat,
    sigma,
    A,
    d,
    theta_grid,
    n_pre_periods,
    post_period_index,
    alpha,
    test_fn,
    **test_kwargs,
):
    """Test whether values in a grid lie in the identified set.

    Parameters
    ----------
    beta_hat : ndarray
        Estimated coefficients.
    sigma : ndarray
        Covariance matrix.
    A : ndarray
        Constraint matrix.
    d : ndarray
        Constraint bounds.
    theta_grid : ndarray
        Grid of theta values to test.
    n_pre_periods : int
        Number of pre-treatment periods.
    post_period_index : int
        Which post-period to test (1-indexed).
    alpha : float
        Significance level.
    test_fn : callable
        Test function to use.
    **test_kwargs
        Additional arguments for test function.

    Returns
    -------
    ndarray
        Array of shape (n_grid, 2) with columns [theta, accept].
    """
    post_period_vec = basis_vector(index=n_pre_periods + post_period_index, size=len(beta_hat)).flatten()

    y_matrix = prepare_theta_grid_y_values(beta_hat, post_period_vec, theta_grid)
    n_grid = len(theta_grid)

    sigma_tilde = np.sqrt(np.diag(A @ sigma @ A.T))
    sigma_tilde = np.maximum(sigma_tilde, 1e-10)
    precomputed = {
        "A_tilde": np.diag(1 / sigma_tilde) @ A,
        "d_tilde": d / sigma_tilde,
    }

    _supports_precomputed = test_fn in (test_in_identified_set, test_in_identified_set_lf_hybrid)

    results = np.zeros((n_grid, 2))
    results[:, 0] = theta_grid
    for i in range(n_grid):
        kwargs = dict(test_kwargs)
        if _supports_precomputed:
            kwargs["_precomputed"] = precomputed
        in_set = test_fn(
            y=y_matrix[i],
            sigma=sigma,
            A=A,
            d=d,
            alpha=alpha,
            **kwargs,
        )
        results[i, 1] = float(in_set)

    return results



"""Functions for computing bounds on the smoothness parameter."""

import numpy as np
from scipy import stats

from .conditional import estimate_lowerbound_m_conditional_test
from .numba import create_bounds_second_difference_matrix, create_monotonicity_matrix


def compute_delta_sd_upperbound_m(
    betahat,
    sigma,
    num_pre_periods,
    alpha=0.05,
):
    r"""Compute an upper bound for :math:`M` at the :math:`1-\alpha` level based on observed pre-period coefficients.

    Constructs an upper bound for the smoothness parameter :math:`M` using the maximum
    second difference of the observed pre-period coefficients.

    The upper bound is computed as the maximum over all second differences of

    .. math::

        \Delta^2 \hat{\beta}_t + z_{1-\alpha} \cdot \text{se}(\Delta^2 \hat{\beta}_t),

    where :math:`\Delta^2 \hat{\beta}_t` is the second difference at time :math:`t`,
    :math:`\text{se}(\cdot)` denotes the standard error, and :math:`z_{1-\alpha}` is the
    :math:`1-\alpha` quantile of the standard normal distribution.

    Parameters
    ----------
    betahat : ndarray
        Vector of estimated coefficients. First `num_pre_periods` elements are pre-treatment.
    sigma : ndarray
        Covariance matrix of estimated coefficients.
    num_pre_periods : int
        Number of pre-treatment periods.
    alpha : float, default=0.05
        Significance level :math:`\alpha` for the confidence bound.

    Returns
    -------
    float
        Upper bound for :math:`M`.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A more credible approach to parallel trends.
        The Review of Economic Studies, 90(5), 2555-2591.
    """
    if num_pre_periods < 3:
        raise ValueError("Cannot estimate M in pre-period with < 3 pre-period coefficients.")

    pre_period_coef = betahat[:num_pre_periods]
    pre_period_sigma = sigma[:num_pre_periods, :num_pre_periods]

    a_sd = create_second_difference_matrix(num_pre_periods=num_pre_periods, num_post_periods=0)

    pre_period_coef_diffs = a_sd @ pre_period_coef
    pre_period_sigma_diffs = a_sd @ pre_period_sigma @ a_sd.T
    se_diffs = np.sqrt(np.diag(pre_period_sigma_diffs))

    upper_bound_vec = pre_period_coef_diffs + stats.norm.ppf(1 - alpha) * se_diffs
    max_upper_bound = np.max(upper_bound_vec)

    return max_upper_bound


def compute_delta_sd_lowerbound_m(
    betahat,
    sigma,
    num_pre_periods,
    alpha=0.05,
    grid_ub=None,
    grid_points=1000,
):
    """Compute a lower bound for :math:`M` using observed pre-period coefficients.

    Constructs a lower bound for the smoothness parameter :math:`M` by constructing a
    one-sided confidence interval on the maximal second difference of the observed
    pre-period coefficients using the conditional test from [1]_.

    Parameters
    ----------
    betahat : ndarray
        Vector of estimated coefficients. First `num_pre_periods` elements are pre-treatment.
    sigma : ndarray
        Covariance matrix of estimated coefficients.
    num_pre_periods : int
        Number of pre-treatment periods.
    alpha : float, default=0.05
        Significance level for the confidence bound.
    grid_ub : float, optional
        Upper bound for grid search. If None, defaults to 3 times the max SE.
    grid_points : int, default=1000
        Number of points in the grid for searching over M values.

    Returns
    -------
    float
        Lower bound for :math:`M`. Returns ``np.inf`` if no values accepted.

    References
    ----------

    .. [1] Andrews, I., Roth, J., & Pakes, A. (2019). Inference for linear conditional moment inequalities.
        Technical report, National Bureau of Economic Research.
    .. [2] Rambachan, A., & Roth, J. (2023). A more credible approach to parallel trends.
        The Review of Economic Studies, 90(5), 2555-2591.
    """
    if num_pre_periods < 3:
        raise ValueError("Cannot estimate M in pre-period with < 3 pre-period coefficients.")

    pre_period_coef = betahat[:num_pre_periods]
    pre_period_sigma = sigma[:num_pre_periods, :num_pre_periods]

    if grid_ub is None:
        grid_ub = 3 * np.max(np.sqrt(np.diag(pre_period_sigma)))

    results = estimate_lowerbound_m_conditional_test(
        pre_period_coef,
        pre_period_sigma,
        grid_ub,
        alpha,
        grid_points,
    )

    return results


def create_second_difference_matrix(
    num_pre_periods,
    num_post_periods,
):
    r"""Create matrix for computing second differences of coefficients.

    Constructs a matrix :math:`A` such that :math:`A \beta` gives the second differences
    of the coefficient vector :math:`\beta`.

    For pre-periods, second differences are computed as

    .. math::

        \Delta^2 \beta_t = \beta_{t+1} - 2 \beta_t + \beta_{t-1}

    for interior points, and

    .. math::

        \Delta^2 \beta_T = \beta_T - 2 \beta_{T-1} + \beta_{T-2}

    for the last pre-period. For post-periods, similar logic applies.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.

    Returns
    -------
    ndarray
        Matrix for computing second differences. Has shape
        :math:`(n_{\text{second_diffs}}, \, n_{\text{pre}} + n_{\text{post}})`.
    """
    return create_bounds_second_difference_matrix(num_pre_periods, num_post_periods)


def create_pre_period_constraint_matrix(num_pre_periods):
    r"""Create constraint matrix and bounds for pre-period second differences.

    Constructs the constraint matrix :math:`A` and bounds :math:`d` such that the constraints
    :math:`|\Delta^2 \beta_i| \leq M` can be written as :math:`A \beta \leq d M`. The constraints are
    set up as :math:`\Delta^2 \beta_i \leq M` for the upper bounds and :math:`-\Delta^2 \beta_i \leq M`
    for the lower bounds.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.

    Returns
    -------
    tuple
        (A, d) where:

        - ``A`` : ndarray
            Constraint matrix of shape :math:`(2 \cdot (n_{\text{pre}} - 1),\ n_{\text{pre}})`
        - ``d`` : ndarray
            Vector of ones of length :math:`2 \cdot (n_{\text{pre}} - 1)`

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A more credible approach to parallel trends.
        The Review of Economic Studies, 90(5), 2555-2591.
    """
    if num_pre_periods < 2:
        raise ValueError("Cannot estimate M in pre-period with < 2 pre-period coefficients.")

    a_tilde = np.zeros((num_pre_periods - 1, num_pre_periods))
    a_tilde[num_pre_periods - 2, (num_pre_periods - 2) : num_pre_periods] = [1, -1]

    if num_pre_periods > 2:
        a_tilde[num_pre_periods - 2, (num_pre_periods - 3) : num_pre_periods] = [1, -2, 1]

        for r in range(num_pre_periods - 3):
            a_tilde[r, r : (r + 3)] = [1, -2, 1]

    a_pre = np.vstack([a_tilde, -a_tilde])
    d = np.ones(a_pre.shape[0])

    return a_pre, d


def create_monotonicity_constraint_matrix(
    num_pre_periods,
    num_post_periods,
    monotonicity_direction="increasing",
    post_period_moments_only=False,
):
    r"""Create constraint matrix for imposing monotonicity restrictions.

    Constructs a matrix :math:`A` such that :math:`A \delta \leq 0` implies
    that :math:`\delta` is monotonic in the specified direction. This is used
    to impose shape restrictions on treatment effect trajectories.

    For an increasing monotonicity constraint, the matrix enforces

    .. math::

        \delta_{t+1} - \delta_t \geq 0 \quad \forall t.

    By constructing :math:`A` such that each row computes :math:`\delta_t - \delta_{t+1}`,
    the constraint :math:`A \delta \leq 0` ensures monotonicity.

    When ``post_period_moments_only=True``, constraints that only involve pre-period
    comparisons are excluded, focusing on monotonicity patterns that involve at
    least one post-period.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    monotonicity_direction : {'increasing', 'decreasing'}, default='increasing'
        Direction of monotonicity constraint.
    post_period_moments_only : bool, default=False
        If True, exclude constraints that only involve pre-periods.

    Returns
    -------
    ndarray
        Constraint matrix :math:`A` of shape :math:`(m, n)` where
        :math:`n = n_{\text{pre}} + n_{\text{post}}` and :math:`m` is the number
        of monotonicity constraints.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A more credible approach to parallel trends.
        The Review of Economic Studies, 90(5), 2555-2591.
    """
    total_periods = num_pre_periods + num_post_periods

    a_m = create_monotonicity_matrix(num_pre_periods, num_post_periods)
    a_m = a_m[~np.all(a_m == 0, axis=1)]

    if post_period_moments_only:
        post_period_indices = list(range(num_pre_periods, total_periods))
        has_post_period = np.any(a_m[:, post_period_indices] != 0, axis=1)
        a_m = a_m[has_post_period]

    if monotonicity_direction == "decreasing":
        a_m = -a_m
    elif monotonicity_direction != "increasing":
        raise ValueError("monotonicity_direction must be 'increasing' or 'decreasing'")

    return a_m


def create_sign_constraint_matrix(
    num_pre_periods,
    num_post_periods,
    bias_direction="positive",
):
    r"""Create constraint matrix for imposing sign restrictions.

    Constructs a matrix :math:`A` such that :math:`A \delta \leq 0` implies
    that the post-period effects have the specified sign.

    For positive bias direction, the constraint enforces

    .. math::

        \delta_t \geq 0 \quad \forall t \in \text{post-periods}.

    For negative bias direction, the constraint enforces

    .. math::

        \delta_t \leq 0 \quad \forall t \in \text{post-periods}.

    This is achieved by setting :math:`A = I` for the post-period coefficients,
    where :math:`I` is the identity matrix, so that :math:`\delta_t \leq 0`.

    Parameters
    ----------
    num_pre_periods : int
        Number of pre-treatment periods.
    num_post_periods : int
        Number of post-treatment periods.
    bias_direction : {'positive', 'negative'}, default='positive'
        Direction of the sign restriction on post-period effects.

    Returns
    -------
    ndarray
        Constraint matrix :math:`A` of shape :math:`(n_{\text{post}}, \, n_{\text{pre}} + n_{\text{post}})`.

    References
    ----------

    .. [1] Rambachan, A., & Roth, J. (2023). A more credible approach to parallel trends.
        The Review of Economic Studies, 90(5), 2555-2591.
    """
    total_periods = num_pre_periods + num_post_periods

    a_b = -np.eye(total_periods)
    a_b = a_b[num_pre_periods : num_pre_periods + num_post_periods, :]

    if bias_direction == "negative":
        a_b = -a_b
    elif bias_direction != "positive":
        raise ValueError("bias_direction must be 'positive' or 'negative'")

    return a_b



"""Entry point for distributed DDD estimation via Spark."""

from __future__ import annotations

import logging

import polars as pl
from pyspark.sql import DataFrame as SparkDataFrame

from moderndid.cupy.backend import _validate_backend_name
from moderndid.didtriple.utils import get_covariate_names

from ._ddd_mp import spark_ddd_mp
from ._utils import get_or_create_spark, validate_spark_input


def spark_ddd(
    data,
    yname,
    tname,
    idname=None,
    gname=None,
    pname=None,
    xformla=None,
    spark=None,
    control_group="nevertreated",
    base_period="universal",
    est_method="dr",
    weightsname=None,
    boot=False,
    biters=1000,
    cband=False,
    cluster=None,
    alpha=0.05,
    trim_level=0.995,
    panel=True,
    allow_unbalanced_panel=False,
    random_state=None,
    n_partitions=None,
    max_cohorts=None,
    backend=None,
):
    r"""Compute the distributed triple difference-in-differences estimator.

    Distributed implementation of the DDD estimator from [1]_ for datasets
    that exceed single-machine memory. This function accepts a PySpark
    DataFrame and dispatches to the distributed multi-period estimator.

    The underlying methodology is identical to :func:`~moderndid.ddd`: for each
    group-time cell :math:`(g,t)`, the estimator computes

    .. math::

        ATT(g,t) = \mathbb{E}[Y_{i,t}(g) - Y_{i,t}(\infty) \mid S_i = g, Q_i = 1],

    where :math:`S_i` is the period when treatment is enabled for unit
    :math:`i`'s group and :math:`Q_i \in \{0,1\}` indicates eligibility.
    Identification relies on a DDD conditional parallel trends assumption that
    allows differential trends between eligible and ineligible units, provided
    these differentials are stable across treatment-enabling groups.

    The distributed backend computes sufficient statistics (Gram matrices,
    cross-products) partition-by-partition and aggregates via tree-reduce,
    avoiding full materialization of the dataset on any single worker.

    Users do not need to call this function directly. Passing a PySpark
    DataFrame to :func:`~moderndid.ddd` will automatically dispatch here.

    Parameters
    ----------
    data : pyspark.sql.DataFrame
        Data in long format as a Spark DataFrame. Each partition should
        contain complete observations (all columns) for a subset of
        units or individuals.
    yname : str
        Name of the outcome variable column.
    tname : str
        Name of the column containing time periods.
    idname : str, optional
        Name of the unit identifier column. Required for panel data.
    gname : str
        Name of the treatment group column. Should be 0 for never-treated
        units and a positive value indicating the first period when treatment
        is enabled for the unit's group.
    pname : str
        Name of the partition/eligibility column (1=eligible, 0=ineligible).
        This identifies which units within a treatment-enabling group are
        actually eligible to receive treatment.
    xformla : str, optional
        Formula for covariates in the form ``"~ x1 + x2"``. If None, only
        an intercept is used.
    spark : pyspark.sql.SparkSession, optional
        Active Spark session. If None, a local session is created
        automatically via :func:`~moderndid.spark.get_or_create_spark`.
    control_group : {"nevertreated", "notyettreated"}, default="nevertreated"
        Which units to use as controls. ``"nevertreated"`` uses only units
        that are never treated. ``"notyettreated"`` additionally includes
        units not yet treated by period :math:`t`.
    base_period : {"universal", "varying"}, default="universal"
        Base period selection for multi-period settings. ``"universal"`` uses
        the period immediately before the first treated period for all
        group-time cells. ``"varying"`` uses the period immediately before
        each group's treatment onset.
    est_method : {"dr", "reg", "ipw"}, default="dr"
        Estimation method. ``"dr"`` uses doubly robust estimation combining
        outcome regression and inverse probability weighting. ``"reg"`` uses
        outcome regression only. ``"ipw"`` uses inverse probability weighting
        only.
    boot : bool, default=False
        Whether to use the multiplier bootstrap for inference.
    biters : int, default=1000
        Number of bootstrap iterations. Only used when ``boot=True``.
    cband : bool, default=False
        Whether to compute uniform confidence bands that cover all group-time
        average treatment effects simultaneously. Only used when
        ``boot=True``.
    cluster : str, optional
        Name of the clustering variable for clustered standard errors.
    alpha : float, default=0.05
        Significance level for confidence intervals.
    random_state : int, optional
        Random seed for reproducibility of bootstrap draws.
    n_partitions : int, optional
        Number of Spark partitions for distributing computation. If None,
        defaults to Spark's default parallelism via
        :func:`~moderndid.spark.get_default_partitions`.
    max_cohorts : int or None, default None
        Maximum number of treatment cohorts to process in parallel.
        When ``None``, defaults to the number of Spark executor cores.

    Returns
    -------
    DDDMultiPeriodResult
        Group-time DDD average treatment effect results containing:

        - **att**: Array of ATT(g,t) point estimates
        - **se**: Standard errors for each ATT(g,t)
        - **uci**, **lci**: Confidence interval bounds
        - **groups**, **times**: Treatment cohort and time for each estimate
        - **glist**, **tlist**: Unique cohorts and periods
        - **inf_func_mat**: Influence function matrix
        - **n**: Number of units
        - **args**: Estimation arguments

    Examples
    --------
    For datasets that fit in memory, create a Spark DataFrame from an existing
    pandas or polars DataFrame:

    .. code-block:: python

        from pyspark.sql import SparkSession
        from moderndid import ddd, gen_dgp_mult_periods

        spark = SparkSession.builder.master("local[*]").getOrCreate()
        dgp = gen_dgp_mult_periods(n=500, dgp_type=1, random_state=42)
        sdf = spark.createDataFrame(dgp["data"].to_pandas())

        result = ddd(
            data=sdf,
            yname="y",
            tname="time",
            idname="id",
            gname="group",
            pname="partition",
            est_method="dr",
        )

    For large datasets stored on disk, read directly into Spark:

    .. code-block:: python

        sdf = spark.read.parquet("large_panel/*.parquet")
        result = ddd(
            data=sdf,
            yname="y",
            tname="time",
            idname="id",
            gname="group",
            pname="partition",
        )

    To use an existing Spark session:

    .. code-block:: python

        from moderndid.spark import spark_ddd

        result = spark_ddd(
            data=sdf,
            yname="y",
            tname="time",
            idname="id",
            gname="group",
            pname="partition",
            spark=spark,
        )

    See Also
    --------
    ddd : Local (non-distributed) DDD estimator. Automatically dispatches
        to ``spark_ddd`` when passed a Spark DataFrame.
    get_or_create_spark : Get or create a Spark session.
    get_default_partitions : Compute default partition count from Spark.

    References
    ----------

    .. [1] Ortiz-Villavicencio, M., & Sant'Anna, P. H. C. (2025).
        *Better Understanding Triple Differences Estimators.*
        arXiv preprint arXiv:2505.09942. https://arxiv.org/abs/2505.09942
    """
    if gname is None:
        raise ValueError("gname is required.")
    if pname is None:
        raise ValueError("pname is required.")

    use_gpu = False
    if backend is not None:
        _validate_backend_name(backend)
        use_gpu = backend.lower() == "cupy"

    spark = get_or_create_spark(spark)
    logging.getLogger("py4j").setLevel(logging.ERROR)

    is_spark = isinstance(data, SparkDataFrame)

    if not is_spark:
        if isinstance(data, pl.DataFrame):
            pdf = data.to_pandas()
            sdf = spark.createDataFrame(pdf)
        else:
            sdf = spark.createDataFrame(data)
    else:
        sdf = data

    required_cols = [yname, tname, gname, pname]
    if idname is not None:
        required_cols.append(idname)
    if weightsname is not None:
        required_cols.append(weightsname)
    validate_spark_input(sdf, required_cols)

    covariate_cols = get_covariate_names(xformla)

    return spark_ddd_mp(
        spark=spark,
        data=sdf,
        y_col=yname,
        time_col=tname,
        id_col=idname,
        group_col=gname,
        partition_col=pname,
        covariate_cols=covariate_cols,
        control_group=control_group,
        base_period=base_period,
        est_method=est_method,
        weightsname=weightsname,
        boot=boot,
        biters=biters,
        cband=cband,
        cluster=cluster,
        alpha=alpha,
        trim_level=trim_level,
        allow_unbalanced_panel=allow_unbalanced_panel,
        random_state=random_state,
        n_partitions=n_partitions,
        max_cohorts=max_cohorts,
        panel=panel,
        use_gpu=use_gpu,
    )



"""Data transformation classes for preprocessing."""

import warnings
from typing import Protocol

import numpy as np
import polars as pl

from ..dataframe import DataFrame, to_polars
from .base import BaseTransformer
from .config import BasePreprocessConfig, ContDIDConfig, DDDConfig, DIDConfig, DIDInterConfig, TwoPeriodDIDConfig
from .constants import (
    NEVER_TREATED_VALUE,
    ROW_ID_COLUMN,
    WEIGHTS_COLUMN,
    ControlGroup,
    DataFormat,
)
from .utils import create_ddd_subgroups, extract_vars_from_formula, make_balanced_panel, validate_subgroup_sizes


class DataTransformer(Protocol):
    """Data transformer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""


class ColumnSelector(BaseTransformer):
    """Column selector."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)
        cols_to_keep = [config.yname, config.tname, config.gname]

        if config.idname:
            cols_to_keep.append(config.idname)

        if config.weightsname:
            cols_to_keep.append(config.weightsname)

        if config.clustervars:
            cols_to_keep.extend(config.clustervars)

        if config.xformla and config.xformla != "~1":
            formula_vars = extract_vars_from_formula(config.xformla)
            formula_vars = [v for v in formula_vars if v != config.yname]
            cols_to_keep.extend(formula_vars)

        if isinstance(config, ContDIDConfig) and config.dname:
            cols_to_keep.append(config.dname)

        cols_to_keep = list(dict.fromkeys(cols_to_keep))
        cols_to_keep = [col for col in cols_to_keep if col is not None]

        return df.select(cols_to_keep)


class MissingDataHandler(BaseTransformer):
    """Missing data."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)

        if isinstance(config, DIDInterConfig):
            df = df.with_columns(
                [
                    pl.col(config.dname).mean().over(config.gname).alias("_mean_D"),
                    pl.col(config.yname).mean().over(config.gname).alias("_mean_Y"),
                ]
            )
            df = df.filter(pl.col("_mean_D").is_not_null() & pl.col("_mean_Y").is_not_null())
            df = df.drop(["_mean_D", "_mean_Y"])
            return df

        n_orig = len(df)
        data_clean = df.drop_nulls()
        n_new = len(data_clean)

        if n_orig > n_new:
            if isinstance(config, TwoPeriodDIDConfig) and config.panel:
                raise ValueError(
                    f"Missing values found in panel data. Dropped {n_orig - n_new} rows. "
                    "Panel data requires complete observations for all time periods. "
                    "Please handle missing values before preprocessing."
                )
            warnings.warn(f"Dropped {n_orig - n_new} rows from original data due to missing values")

        return data_clean


class WeightNormalizer(BaseTransformer):
    """Weight normalizer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)

        weights = np.ones(len(df)) if config.weightsname is None else df[config.weightsname].to_numpy()

        weights = weights / weights.mean()
        return df.with_columns(pl.Series(name=WEIGHTS_COLUMN, values=weights))


class DataSorter(BaseTransformer):
    """Data sorter."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)
        sort_cols = [config.tname, config.gname]

        idname = getattr(config, "idname", None)
        if idname:
            sort_cols.append(idname)

        return df.sort(sort_cols)


class TreatmentEncoder(BaseTransformer):
    """Treatment encoder."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)

        df = df.with_columns(pl.col(config.gname).cast(pl.Float64))

        df = df.with_columns(
            pl.when(pl.col(config.gname) == 0)
            .then(pl.lit(NEVER_TREATED_VALUE))
            .otherwise(pl.col(config.gname))
            .alias(config.gname)
        )

        tlist = sorted(df[config.tname].unique().to_list())
        max_treatment_time = max(tlist)
        df = df.with_columns(
            pl.when(pl.col(config.gname) > max_treatment_time)
            .then(pl.lit(NEVER_TREATED_VALUE))
            .otherwise(pl.col(config.gname))
            .alias(config.gname)
        )

        return df


class EarlyTreatmentFilter(BaseTransformer):
    """Early treatment filter."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)

        tlist = sorted(df[config.tname].unique().to_list())
        first_period = min(tlist)

        treated_early_mask = pl.col(config.gname) <= first_period + config.anticipation

        if config.idname:
            early_units = df.filter(treated_early_mask)[config.idname].unique().to_list()
            n_early = len(early_units)
            if n_early > 0:
                warnings.warn(f"Dropped {n_early} units that were already treated in the first period")
                df = df.filter(~pl.col(config.idname).is_in(early_units))
        else:
            n_early = df.filter(treated_early_mask).height
            if n_early > 0:
                warnings.warn(f"Dropped {n_early} observations that were already treated in the first period")
                df = df.filter(~treated_early_mask)

        return df


class ControlGroupCreator(BaseTransformer):
    """Control group creator."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDConfig):
            return to_polars(data)

        df = to_polars(data)

        glist = sorted(df[config.gname].unique().to_list())

        if NEVER_TREATED_VALUE in glist:
            return df

        finite_glist = [g for g in glist if np.isfinite(g)]
        if not finite_glist:
            return df

        latest_g = max(finite_glist)
        cutoff_t = latest_g - config.anticipation

        if config.control_group == ControlGroup.NEVER_TREATED:
            warnings.warn(
                "No never-treated group is available. "
                "The last treated cohort is being coerced as 'never-treated' units."
            )
            df = df.filter(pl.col(config.tname) < cutoff_t)
            df = df.with_columns(
                pl.when(pl.col(config.gname) == latest_g)
                .then(pl.lit(NEVER_TREATED_VALUE))
                .otherwise(pl.col(config.gname))
                .alias(config.gname)
            )
        else:
            df = df.filter(pl.col(config.tname) < cutoff_t)

        return df


class PanelBalancer(BaseTransformer):
    """Panel balancer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not config.panel or config.allow_unbalanced_panel or not config.idname:
            return to_polars(data)

        df = to_polars(data)
        tlist = sorted(df[config.tname].unique().to_list())
        n_periods = len(tlist)

        unit_counts = df.group_by(config.idname).len()
        complete_units = unit_counts.filter(pl.col("len") == n_periods)[config.idname].to_list()

        n_old = df[config.idname].n_unique()
        df = df.filter(pl.col(config.idname).is_in(complete_units))
        n_new = df[config.idname].n_unique()

        if n_new < n_old:
            warnings.warn(f"Dropped {n_old - n_new} units while converting to balanced panel")

        if len(df) == 0:
            raise ValueError(
                "All observations dropped while converting to balanced panel. "
                "Consider setting panel=False and/or revisiting 'idname'"
            )

        return df


class RepeatedCrossSectionHandler(BaseTransformer):
    """Repeated cross section handler."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if config.panel:
            return to_polars(data)

        df = to_polars(data)

        if config.idname is None:
            config.true_repeated_cross_sections = True

        if config.true_repeated_cross_sections:
            df = df.with_row_index(name=ROW_ID_COLUMN)
            config.idname = ROW_ID_COLUMN
        else:
            df = df.with_columns(pl.col(config.idname).alias(ROW_ID_COLUMN))

        return df


class TimePeriodRecoder(BaseTransformer):
    """Time period recoder."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, ContDIDConfig):
            return to_polars(data)

        df = to_polars(data)
        original_periods = sorted(df[config.tname].unique().to_list())
        time_map = {t: i + 1 for i, t in enumerate(original_periods)}

        df = df.with_columns(pl.col(config.tname).replace(time_map).alias(config.tname))
        config.time_map = time_map

        return df


class EarlyTreatmentGroupFilter(BaseTransformer):
    """Early treatment group filter."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, ContDIDConfig):
            return to_polars(data)

        df = to_polars(data)

        glist = sorted([g for g in df[config.gname].unique().to_list() if np.isfinite(g)])
        tlist = sorted(df[config.tname].unique().to_list())

        if not glist:
            return df

        min_valid_group = config.required_pre_periods + config.anticipation + min(tlist)

        groups_to_drop = [g for g in glist if g < min_valid_group]

        if groups_to_drop:
            warnings.warn(
                f"Dropped {len(groups_to_drop)} groups treated before period {min_valid_group} "
                f"(required_pre_periods={config.required_pre_periods}, anticipation={config.anticipation})"
            )
            df = df.filter(~pl.col(config.gname).is_in(groups_to_drop))

        return df


class DoseValidatorTransformer(BaseTransformer):
    """Dose validator transformer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, ContDIDConfig) or not config.dname:
            return to_polars(data)

        df = to_polars(data)

        df = df.with_columns(
            pl.when(pl.col(config.gname) == NEVER_TREATED_VALUE)
            .then(pl.lit(0))
            .otherwise(pl.col(config.dname))
            .alias(config.dname)
        )

        invalid_mask = (
            (pl.col(config.gname) != NEVER_TREATED_VALUE)
            & pl.col(config.gname).is_finite()
            & (pl.col(config.tname) >= pl.col(config.gname))
            & ((pl.col(config.dname) == 0) | pl.col(config.dname).is_null())
        )

        n_invalid = df.filter(invalid_mask).height

        if n_invalid > 0:
            warnings.warn(f"Dropped {n_invalid} post-treatment observations with missing or zero dose values")
            df = df.filter(~invalid_mask)

        return df


class ConfigUpdater:
    """Config updater."""

    @staticmethod
    def update(data: DataFrame, config: BasePreprocessConfig) -> None:
        """Update config."""
        df = to_polars(data)

        if isinstance(config, TwoPeriodDIDConfig):
            tlist = sorted(df[config.tname].unique().to_list())
            treat_list = sorted(df[config.treat_col].unique().to_list())

            n_units = df[config.idname].n_unique() if config.idname else len(df)

            config.time_periods = np.array(tlist)
            config.time_periods_count = len(tlist)
            config.treated_groups = np.array(treat_list)
            config.treated_groups_count = len(treat_list)
            config.id_count = n_units
            return

        tlist = sorted(df[config.tname].unique().to_list())
        glist = sorted(df[config.gname].unique().to_list())

        glist_finite = [g for g in glist if np.isfinite(g)]

        n_units = df[config.idname].n_unique() if config.idname else len(df)

        config.time_periods = np.array(tlist)
        config.time_periods_count = len(tlist)
        config.treated_groups = np.array(glist_finite)
        config.treated_groups_count = len(glist_finite)
        config.id_count = n_units

        if config.panel and config.allow_unbalanced_panel:
            unit_counts = df.group_by(config.idname).len()
            is_balanced = (unit_counts["len"] == len(tlist)).all()
            if is_balanced:
                config.data_format = DataFormat.PANEL
            else:
                config.data_format = DataFormat.UNBALANCED_PANEL
        elif config.panel:
            config.data_format = DataFormat.PANEL
        else:
            config.data_format = DataFormat.REPEATED_CROSS_SECTION

        if len(tlist) == 2:
            config.cband = False


class PrePostColumnSelector(BaseTransformer):
    """Pre-post column selector."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig | TwoPeriodDIDConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, TwoPeriodDIDConfig):
            return to_polars(data)

        df = to_polars(data)
        cols_to_keep = [config.yname, config.tname, config.treat_col]

        if config.idname:
            cols_to_keep.append(config.idname)

        if config.weightsname:
            cols_to_keep.append(config.weightsname)

        if config.xformla and config.xformla != "~1":
            formula_vars = extract_vars_from_formula(config.xformla)
            formula_vars = [v for v in formula_vars if v != config.yname]
            cols_to_keep.extend(formula_vars)

        cols_to_keep = list(dict.fromkeys(cols_to_keep))
        cols_to_keep = [col for col in cols_to_keep if col is not None]

        return df.select(cols_to_keep)


class PrePostCovariateProcessor(BaseTransformer):
    """Pre-post covariate processor."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig | TwoPeriodDIDConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, TwoPeriodDIDConfig):
            return to_polars(data)

        import formulaic as fml

        df = to_polars(data)
        covariates_formula = config.xformla if config.xformla else "~1"

        try:
            model_matrix_result = fml.model_matrix(covariates_formula, df)
            covariates_pl = model_matrix_result.__wrapped__

            if hasattr(model_matrix_result, "model_spec") and model_matrix_result.model_spec:
                original_cov_names = [var for var in model_matrix_result.model_spec.variables if var != "1"]
            else:
                original_cov_names = []
                warnings.warn("Could not retrieve model_spec from formulaic output.", UserWarning)

        except Exception as e:
            raise ValueError(f"Error processing covariates_formula '{covariates_formula}' with formulaic: {e}") from e

        cols_to_drop = [name for name in original_cov_names if name in df.columns]
        cols_to_keep = [col for col in df.columns if col not in cols_to_drop]

        return pl.concat([df.select(cols_to_keep), covariates_pl], how="horizontal")


class PrePostPanelBalancer(BaseTransformer):
    """Pre-post panel balancer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig | TwoPeriodDIDConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, TwoPeriodDIDConfig) or not config.panel or not config.idname:
            return to_polars(data)

        df = to_polars(data)
        n_times = df[config.tname].n_unique()
        obs_counts = df.group_by(config.idname).len()
        ids_to_keep = obs_counts.filter(pl.col("len") == n_times)[config.idname].to_list()

        if len(ids_to_keep) < obs_counts.height:
            warnings.warn("Panel data is unbalanced. Dropping units with incomplete observations.", UserWarning)

        return df.filter(pl.col(config.idname).is_in(ids_to_keep))


class PrePostInvarianceChecker(BaseTransformer):
    """Pre-post invariance checker."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig | TwoPeriodDIDConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, TwoPeriodDIDConfig) or not config.panel or not config.idname:
            return to_polars(data)

        df = to_polars(data)
        time_periods = sorted(df[config.tname].unique().to_list())
        if len(time_periods) != 2:
            return df

        pre_period, post_period = time_periods

        pre_df = df.filter(pl.col(config.tname) == pre_period).sort(config.idname)
        post_df = df.filter(pl.col(config.tname) == post_period).sort(config.idname)

        pre_ids = set(pre_df[config.idname].to_list())
        post_ids = set(post_df[config.idname].to_list())
        common_ids = list(pre_ids.intersection(post_ids))

        pre_df = pre_df.filter(pl.col(config.idname).is_in(common_ids)).sort(config.idname)
        post_df = post_df.filter(pl.col(config.idname).is_in(common_ids)).sort(config.idname)

        if not pre_df[config.treat_col].equals(post_df[config.treat_col]):
            raise ValueError(f"Treatment indicator ('{config.treat_col}') must be time-invariant in panel data.")

        if (
            WEIGHTS_COLUMN in pre_df.columns
            and WEIGHTS_COLUMN in post_df.columns
            and not pre_df[WEIGHTS_COLUMN].equals(post_df[WEIGHTS_COLUMN])
        ):
            raise ValueError("Weights must be time-invariant in panel data.")

        return df


class DIDInterColumnSelector(BaseTransformer):
    """DIDInter column selector."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)
        cols_to_keep = [config.yname, config.tname, config.gname, config.dname]

        if config.weightsname:
            cols_to_keep.append(config.weightsname)

        if config.cluster:
            cols_to_keep.append(config.cluster)

        if config.xformla and config.xformla != "~1":
            covariate_names = extract_vars_from_formula(config.xformla)
            cols_to_keep.extend(covariate_names)

        if config.trends_nonparam:
            cols_to_keep.extend(config.trends_nonparam)

        if config.predict_het:
            cols_to_keep.extend(config.predict_het[0])

        cols_to_keep = list(dict.fromkeys(cols_to_keep))
        cols_to_keep = [col for col in cols_to_keep if col is not None and col in df.columns]

        df = df.select(cols_to_keep)

        for col in [config.tname, config.gname, config.dname]:
            if col in df.columns and df[col].dtype not in (pl.Float64, pl.Float32, pl.Int64, pl.Int32):
                df = df.with_columns(pl.col(col).cast(pl.Float64))

        return df


class SwitcherIdentifier(BaseTransformer):
    """Identify switchers."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)
        df = df.sort([config.gname, config.tname])

        df = df.with_columns((pl.col(config.dname) - pl.col(config.dname).shift(1).over(config.gname)).alias("_d_diff"))

        base_treatment_pre = (
            df.filter(pl.col(config.tname) == pl.col(config.tname).min().over(config.gname))
            .select([config.gname, pl.col(config.dname).alias("_d_sq_pre")])
            .unique()
        )
        df = df.join(base_treatment_pre, on=config.gname, how="left")
        df = df.with_columns((pl.col(config.dname) - pl.col("_d_sq_pre")).alias("_diff_from_sq"))

        first_switch_pre = (
            df.filter((pl.col("_d_diff") != 0) & pl.col("_d_diff").is_not_null())
            .group_by(config.gname)
            .agg(pl.col(config.tname).min().alias("_F_g_pre"))
        )
        df = df.join(first_switch_pre, on=config.gname, how="left")

        t_max_per_unit = df.group_by(config.gname).agg(pl.col(config.tname).max().alias("_T_max_unit"))
        df = df.join(t_max_per_unit, on=config.gname, how="left")

        df = df.with_columns(
            pl.when(pl.col("_F_g_pre").is_not_null())
            .then(pl.col("_T_max_unit") - pl.col("_F_g_pre") + 1)
            .otherwise(pl.lit(0.0))
            .alias("L_g")
        )
        df = df.drop(["_F_g_pre", "_T_max_unit"])

        df = df.with_columns(
            [
                pl.when((pl.col("_diff_from_sq") > 0) & pl.col(config.dname).is_not_null())
                .then(1)
                .otherwise(0)
                .cum_sum()
                .clip(upper_bound=1)
                .over(config.gname)
                .alias("_ever_strict_increase"),
                pl.when((pl.col("_diff_from_sq") < 0) & pl.col(config.dname).is_not_null())
                .then(1)
                .otherwise(0)
                .cum_sum()
                .clip(upper_bound=1)
                .over(config.gname)
                .alias("_ever_strict_decrease"),
            ]
        )

        if not config.keep_bidirectional_switchers:
            df = df.filter(~((pl.col("_ever_strict_increase") == 1) & (pl.col("_ever_strict_decrease") == 1)))

        df = df.drop(["_ever_strict_increase", "_ever_strict_decrease", "_d_sq_pre", "_diff_from_sq"])

        first_switch = (
            df.filter((pl.col("_d_diff") != 0) & pl.col("_d_diff").is_not_null())
            .group_by(config.gname)
            .agg(pl.col(config.tname).min().alias("F_g"))
        )

        df = df.join(first_switch, on=config.gname, how="left")
        df = df.with_columns(pl.col("F_g").fill_null(float("inf")))

        base_treatment = (
            df.filter(pl.col(config.tname) == pl.col(config.tname).min().over(config.gname))
            .select([config.gname, pl.col(config.dname).alias("d_sq")])
            .unique()
        )
        df = df.join(base_treatment, on=config.gname, how="left")

        switch_treatment = (
            df.filter(pl.col(config.tname) == pl.col("F_g"))
            .select([config.gname, pl.col(config.dname).alias("d_fg")])
            .unique()
        )
        df = df.join(switch_treatment, on=config.gname, how="left")

        switch_direction = (
            df.filter(pl.col("_d_diff").is_not_null() & (pl.col("_d_diff") != 0))
            .group_by(config.gname)
            .agg(pl.col("_d_diff").first().alias("_first_diff"))
        )
        df = df.join(switch_direction, on=config.gname, how="left")

        df = df.with_columns(
            pl.when(pl.col("F_g") == float("inf"))
            .then(0)
            .when(pl.col("_first_diff") > 0)
            .then(1)
            .when(pl.col("_first_diff") < 0)
            .then(-1)
            .otherwise(0)
            .alias("S_g")
        )

        if config.drop_missing_preswitch:
            min_treat_time = (
                df.filter(pl.col(config.dname).is_not_null())
                .group_by(config.gname)
                .agg(pl.col(config.tname).min().alias("_min_treat_time"))
            )
            df = df.join(min_treat_time, on=config.gname, how="left")

            df = df.filter(
                ~(
                    (pl.col("_min_treat_time") < pl.col("F_g"))
                    & (pl.col(config.tname) >= pl.col("_min_treat_time"))
                    & (pl.col(config.tname) < pl.col("F_g"))
                    & pl.col(config.dname).is_null()
                )
            )
            df = df.drop("_min_treat_time")

        df = df.drop(["_d_diff", "_first_diff"])

        return df


class SwitcherFilter(BaseTransformer):
    """Filter units based on switchers parameter."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)

        if config.switchers == "in":
            valid_units = df.filter((pl.col("S_g") == 1) | (pl.col("S_g") == 0))[config.gname].unique().to_list()
            df = df.filter(pl.col(config.gname).is_in(valid_units))
        elif config.switchers == "out":
            valid_units = df.filter((pl.col("S_g") == -1) | (pl.col("S_g") == 0))[config.gname].unique().to_list()
            df = df.filter(pl.col(config.gname).is_in(valid_units))

        return df


class FgVariationFilter(BaseTransformer):
    """Filter out baseline treatment groups with no variation in F_g."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)

        group_cols = ["d_sq"]
        if config.trends_nonparam:
            group_cols.extend(config.trends_nonparam)

        df = df.with_columns(
            pl.when(pl.col("F_g") == float("inf")).then(0).otherwise(pl.col("F_g")).alias("_F_g_for_std")
        )

        df = df.with_columns(pl.col("_F_g_for_std").std().over(group_cols).round(3).alias("_var_F_g"))

        df = df.filter(pl.col("_var_F_g") > 0)
        df = df.drop(["_var_F_g", "_F_g_for_std"])

        return df


class ControlsTimeFilter(BaseTransformer):
    """Filter to cells with at least one never-switcher as control."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)

        df = df.with_columns(pl.when(pl.col("F_g") == float("inf")).then(1).otherwise(0).alias("_never_change_d"))

        ctrl_group = [config.tname, "d_sq"]
        if config.trends_nonparam:
            ctrl_group.extend(config.trends_nonparam)

        df = df.with_columns(pl.col("_never_change_d").max().over(ctrl_group).alias("_controls_time"))

        df = df.filter(pl.col("_controls_time") > 0)
        df = df.drop(["_never_change_d", "_controls_time"])

        return df


class ContinuousTreatmentProcessor(BaseTransformer):
    """Process continuous treatment for DIDInter."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        if config.continuous <= 0:
            return to_polars(data)

        df = to_polars(data)
        degree_pol = config.continuous

        df = df.with_columns(pl.col("d_sq").alias("d_sq_orig"))

        for p in range(1, degree_pol + 1):
            df = df.with_columns((pl.col("d_sq_orig") ** p).alias(f"d_sq_{p}"))

        df = df.with_columns(pl.lit(0).alias("d_sq"))

        mapping_df = (
            df.select("d_sq")
            .filter(pl.col("d_sq").is_not_null())
            .unique()
            .sort("d_sq")
            .with_row_index("d_sq_int", offset=1)
            .select(["d_sq", "d_sq_int"])
        )
        df = df.join(mapping_df, on="d_sq", how="left")

        return df


class DIDInterPanelBalancer(BaseTransformer):
    """DIDInter panel balancer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)

        groups = df.select(config.gname).unique()
        times = df.select(config.tname).unique()

        full_index = groups.join(times, how="cross")

        df = full_index.join(df, on=[config.gname, config.tname], how="left")

        time_invariant_cols = ["F_g", "d_sq", "S_g", "L_g"]
        for col in time_invariant_cols:
            if col in df.columns:
                df = df.with_columns(pl.col(col).mean().over(config.gname).alias(col))

        return df


class DIDInterConfigUpdater:
    """DIDInter config updater."""

    @staticmethod
    def update(data: DataFrame, config: DIDInterConfig) -> None:
        """Update config."""
        df = to_polars(data)

        tlist = sorted(df[config.tname].unique().to_list())
        n_groups = df[config.gname].n_unique()

        config.time_periods = np.array(tlist)
        config.time_periods_count = len(tlist)
        config.n_groups = n_groups
        config.id_count = n_groups

        T_max = int(df[config.tname].max())
        T_min = int(df[config.tname].min())

        switchers = df.filter(pl.col("F_g") != float("inf"))
        if len(switchers) > 0:
            max_effects = (
                switchers.group_by(config.gname)
                .agg((pl.lit(T_max) - pl.col("F_g") + 1).max().alias("max_exp"))
                .select("max_exp")
                .max()
                .item()
            )
            max_placebo = (
                switchers.group_by(config.gname)
                .agg((pl.col("F_g") - pl.lit(T_min) - 1).max().alias("max_pre"))
                .select("max_pre")
                .max()
                .item()
            )
            config.max_effects_available = int(max_effects) if max_effects is not None else 0
            config.max_placebo_available = int(max_placebo) if max_placebo is not None else 0
        else:
            config.max_effects_available = 0
            config.max_placebo_available = 0

        if config.effects > config.max_effects_available:
            warnings.warn(
                f"Requested effects={config.effects} but only {config.max_effects_available} "
                f"post-treatment periods available. Using effects={config.max_effects_available}.",
                UserWarning,
                stacklevel=4,
            )
            config.effects = config.max_effects_available

        if config.placebo > config.max_placebo_available:
            warnings.warn(
                f"Requested placebo={config.placebo} but only {config.max_placebo_available} "
                f"pre-treatment periods available. Using placebo={config.max_placebo_available}.",
                UserWarning,
                stacklevel=4,
            )
            config.placebo = config.max_placebo_available

        if config.allow_unbalanced_panel:
            unit_counts = df.group_by(config.gname).len()
            is_balanced = (unit_counts["len"] == len(tlist)).all()
            if is_balanced:
                config.data_format = DataFormat.PANEL
            else:
                config.data_format = DataFormat.UNBALANCED_PANEL
        else:
            config.data_format = DataFormat.PANEL


class TrendsLinTransformer(BaseTransformer):
    """Apply first-differencing transformation for linear trends."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)

        if not config.trends_lin:
            return df

        t_min = df[config.tname].min()
        df = df.filter(pl.col("F_g") != t_min + 1)
        df = df.sort([config.gname, config.tname])

        df = df.with_columns(
            (pl.col(config.yname) - pl.col(config.yname).shift(1).over(config.gname)).alias(config.yname)
        )

        if config.xformla and config.xformla != "~1":
            covariate_names = extract_vars_from_formula(config.xformla)
            for ctrl in covariate_names:
                df = df.with_columns((pl.col(ctrl) - pl.col(ctrl).shift(1).over(config.gname)).alias(ctrl))

        df = df.filter(pl.col(config.tname) != t_min)

        return df


class DIDInterDataPreparer(BaseTransformer):
    """Prepare data for DIDInter computation."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        if not isinstance(config, DIDInterConfig):
            return to_polars(data)

        df = to_polars(data)
        gname = config.gname
        tname = config.tname
        yname = config.yname
        dname = config.dname

        df = df.sort([gname, tname])
        df = df.with_columns(pl.lit(1.0).alias("weight_gt"))

        if config.weightsname:
            df = df.with_columns(pl.col(config.weightsname).fill_null(0).alias("weight_gt"))

        df = df.with_columns(
            pl.when(pl.col(yname).is_null() | pl.col(dname).is_null())
            .then(0.0)
            .otherwise(pl.col("weight_gt"))
            .alias("weight_gt")
        )

        first_obs = df.group_by(gname).agg(pl.col(tname).min().alias("_first_t")).select([gname, "_first_t"])
        df = df.join(first_obs, on=gname, how="left")
        df = df.with_columns((pl.col(tname) == pl.col("_first_t")).cast(pl.Int64).alias("first_obs_by_gp"))
        df = df.drop("_first_t")

        t_max_by_group = df.group_by(gname).agg(pl.col(tname).max().alias("t_max_by_group"))
        df = df.join(t_max_by_group, on=gname, how="left")

        group_cols = ["d_sq"]
        if config.trends_nonparam:
            group_cols.extend(config.trends_nonparam)

        df = df.with_columns(
            pl.when(pl.col("F_g") == float("inf"))
            .then(pl.col("t_max_by_group") + 1)
            .otherwise(pl.col("F_g"))
            .alias("_F_g_trunc")
        )
        df = df.with_columns((pl.col("_F_g_trunc").max().over(group_cols) - 1).alias("T_g"))
        df = df.drop("_F_g_trunc")

        return df


class DDDColumnSelector(BaseTransformer):
    """DDD column selector."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Select relevant columns for DDD preprocessing."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)
        cols_to_keep = [config.yname, config.tname, config.idname, config.gname, config.pname]

        if config.cluster:
            cols_to_keep.append(config.cluster)

        if config.weightsname:
            cols_to_keep.append(config.weightsname)

        if config.xformla and config.xformla != "~1":
            formula_vars = extract_vars_from_formula(config.xformla)
            cols_to_keep.extend(formula_vars)

        cols_to_keep = list(dict.fromkeys(cols_to_keep))
        cols_to_keep = [col for col in cols_to_keep if col is not None and col in df.columns]

        return df.select(cols_to_keep)


class DDDWeightProcessor(BaseTransformer):
    """DDD weight processor."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Extract/create weights, validate, normalize, add as WEIGHTS_COLUMN."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)

        if config.weightsname is not None:
            weights = df[config.weightsname].to_numpy().astype(float)
            if np.any(np.isnan(weights)):
                raise ValueError("Missing values in weights column.")
            weights_per_id = df.group_by(config.idname).agg(pl.col(config.weightsname).n_unique().alias("n_unique"))
            if (weights_per_id["n_unique"] > 1).any():
                raise ValueError("Weights must be the same across all periods for each unit.")
        else:
            weights = np.ones(len(df))

        weights = weights / np.mean(weights)
        return df.with_columns(pl.Series(name=WEIGHTS_COLUMN, values=weights))


class DDDPanelBalancer(BaseTransformer):
    """DDD panel balancer."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Sort and balance the panel."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)
        df = df.sort([config.idname, config.tname])
        df = make_balanced_panel(df, config.idname, config.tname)

        if len(df) == 0:
            raise ValueError("No observations remain after creating balanced panel.")

        return df


class DDDPostIndicatorCreator(BaseTransformer):
    """DDD post indicator creator."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Create _post column indicating post-treatment period."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)
        tlist = np.sort(df[config.tname].unique().to_numpy())
        return df.with_columns((pl.col(config.tname) == tlist[1]).cast(pl.Int64).alias("_post"))


class DDDCovariateInvarianceChecker(BaseTransformer):
    """DDD covariate invariance checker."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Check that covariates are time-invariant."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)

        if config.xformla == "~1":
            return df

        covariate_vars = extract_vars_from_formula(config.xformla)

        for var in covariate_vars:
            if var not in df.columns:
                continue
            var_per_id = df.group_by(config.idname).agg(pl.col(var).n_unique().alias("n_unique"))
            if (var_per_id["n_unique"] > 1).any():
                raise ValueError(f"Covariate '{var}' varies over time. Covariates must be time-invariant.")

        return df


class DDDSubgroupCreator(BaseTransformer):
    """DDD subgroup creator."""

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Assign 4-group subgroups and validate sizes."""
        if not isinstance(config, DDDConfig):
            return to_polars(data)

        df = to_polars(data)
        glist = np.sort(df[config.gname].unique().to_numpy())
        treat_val = glist[1]

        subgroup = create_ddd_subgroups(df[config.gname].to_numpy(), df[config.pname].to_numpy(), treat_val)
        df = df.with_columns(pl.Series("_subgroup", subgroup))

        counts_df = df.group_by("_subgroup").agg(pl.col(config.idname).n_unique().alias("count"))
        subgroup_counts = {int(row["_subgroup"]): int(row["count"]) for row in counts_df.iter_rows(named=True)}
        validate_subgroup_sizes(subgroup_counts)

        return df


class DDDConfigUpdater:
    """DDD config updater."""

    @staticmethod
    def update(data: DataFrame, config: DDDConfig) -> None:
        """Update DDD config with computed values."""
        df = to_polars(data)

        tlist = np.sort(df[config.tname].unique().to_numpy())
        config.time_periods = tlist
        config.time_periods_count = len(tlist)

        n_units = df.filter(pl.col("_post") == 0).height
        config.n_units = n_units


class DataTransformerPipeline:
    """Data transformer pipeline."""

    def __init__(self, transformers: list[BaseTransformer] | None = None):
        """Initialize data transformer pipeline."""
        self.transformers = transformers or []

    @staticmethod
    def get_did_pipeline() -> "DataTransformerPipeline":
        """Get DID pipeline."""
        return DataTransformerPipeline(
            [
                ColumnSelector(),
                MissingDataHandler(),
                WeightNormalizer(),
                TreatmentEncoder(),
                EarlyTreatmentFilter(),
                ControlGroupCreator(),
                PanelBalancer(),
                RepeatedCrossSectionHandler(),
                DataSorter(),
            ]
        )

    @staticmethod
    def get_cont_did_pipeline() -> "DataTransformerPipeline":
        """Get ContDID pipeline."""
        return DataTransformerPipeline(
            [
                ColumnSelector(),
                MissingDataHandler(),
                WeightNormalizer(),
                TreatmentEncoder(),
                TimePeriodRecoder(),
                EarlyTreatmentGroupFilter(),
                DoseValidatorTransformer(),
                PanelBalancer(),
                DataSorter(),
            ]
        )

    @staticmethod
    def get_two_period_pipeline() -> "DataTransformerPipeline":
        """Get two-period pipeline."""
        return DataTransformerPipeline(
            [
                PrePostColumnSelector(),
                MissingDataHandler(),
                PrePostCovariateProcessor(),
                WeightNormalizer(),
                PrePostPanelBalancer(),
                PrePostInvarianceChecker(),
            ]
        )

    @staticmethod
    def get_didinter_pipeline() -> "DataTransformerPipeline":
        """Get DIDInter pipeline."""
        return DataTransformerPipeline(
            [
                DIDInterColumnSelector(),
                MissingDataHandler(),
                WeightNormalizer(),
                SwitcherIdentifier(),
                SwitcherFilter(),
                FgVariationFilter(),
                ControlsTimeFilter(),
                ContinuousTreatmentProcessor(),
                DIDInterPanelBalancer(),
                TrendsLinTransformer(),
                DIDInterDataPreparer(),
                DataSorter(),
            ]
        )

    @staticmethod
    def get_ddd_pipeline() -> "DataTransformerPipeline":
        """Get DDD pipeline."""
        return DataTransformerPipeline(
            [
                DDDColumnSelector(),
                MissingDataHandler(),
                DDDWeightProcessor(),
                DDDPanelBalancer(),
                DDDPostIndicatorCreator(),
                DDDCovariateInvarianceChecker(),
                DDDSubgroupCreator(),
            ]
        )

    def transform(self, data: DataFrame, config: BasePreprocessConfig) -> pl.DataFrame:
        """Transform data."""
        df = to_polars(data)
        for transformer in self.transformers:
            df = transformer.transform(df, config)

        if isinstance(config, DDDConfig):
            DDDConfigUpdater.update(df, config)
        elif isinstance(config, DIDInterConfig):
            DIDInterConfigUpdater.update(df, config)
        else:
            ConfigUpdater.update(df, config)

        return df



"""Bootstrap inference for repeated cross-section DiD estimators using LogisticRegression."""

import warnings

import numpy as np
import statsmodels.api as sm

from moderndid.cupy.backend import get_backend, to_numpy
from moderndid.cupy.regression import cupy_logistic_irls

from ..estimators.wols import wols_rc
from ..propensity.aipw_estimators import aipw_did_rc_imp1, aipw_did_rc_imp2
from ..utils import _validate_inputs


def wboot_drdid_rc1(y, post, d, x, i_weights, n_bootstrap=1000, trim_level=0.995, random_state=None):
    r"""Compute bootstrap estimates for control-only doubly-robust DiD with repeated cross-sections.

    Implements the bootstrap inference for the doubly-robust difference-in-differences
    estimator with repeated cross-section data. This version uses outcome regression on control
    units only and standard logistic regression for propensity scores.

    Parameters
    ----------
    y : ndarray
        A 1D array representing the outcome variable for each unit.
    post : ndarray
        A 1D array representing the post-treatment period indicator (1 for post, 0 for pre)
        for each unit.
    d : ndarray
        A 1D array representing the treatment indicator (1 for treated, 0 for control)
        for each unit.
    x : ndarray
        A 2D array of covariates (including intercept if desired) with shape (n_units, n_features).
    i_weights : ndarray
        A 1D array of individual observation weights for each unit.
    n_bootstrap : int
        Number of bootstrap iterations. Default is 1000.
    trim_level : float
        Maximum propensity score value for control units to avoid extreme weights.
        Default is 0.995.
    random_state : int, RandomState instance or None
        Controls the random number generation for reproducibility.

    Returns
    -------
    ndarray
        A 1D array of bootstrap ATT estimates with length n_bootstrap.

    See Also
    --------
    aipw_did_rc_imp1 : The underlying simplified AIPW estimator for repeated cross-sections.
    boot_drdid_rc : Traditional bootstrap for doubly-robust DiD.
    """
    n_units = _validate_inputs({"y": y, "post": post, "d": d, "i_weights": i_weights}, x, n_bootstrap, trim_level)

    rng = np.random.RandomState(random_state)
    bootstrap_estimates = np.zeros(n_bootstrap)

    for b in range(n_bootstrap):
        v = rng.exponential(scale=1.0, size=n_units)
        b_weights = i_weights * v

        try:
            ps_b = _fit_logistic_boot_rc(d, x, b_weights)
        except (ValueError, np.linalg.LinAlgError) as e:
            warnings.warn(f"Propensity score estimation failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan
            continue

        ps_b = np.clip(ps_b, 1e-6, 1 - 1e-6)

        trim_ps = np.ones_like(ps_b, dtype=bool)
        control_mask = d == 0
        trim_ps[control_mask] = ps_b[control_mask] < trim_level

        try:
            control_pre_results = wols_rc(y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=True, treat=False)
            out_y_cont_pre_b = control_pre_results.out_reg

            control_post_results = wols_rc(
                y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=False, treat=False
            )
            out_y_cont_post_b = control_post_results.out_reg

            out_y_b = post * out_y_cont_post_b + (1 - post) * out_y_cont_pre_b

        except (ValueError, np.linalg.LinAlgError, KeyError) as e:
            warnings.warn(f"Outcome regression failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan
            continue

        b_weights_trimmed = b_weights.copy()
        b_weights_trimmed[~trim_ps] = 0

        try:
            att_b = aipw_did_rc_imp1(
                y=y,
                post=post,
                d=d,
                ps=ps_b,
                out_reg=out_y_b,
                i_weights=b_weights_trimmed,
            )
            bootstrap_estimates[b] = att_b
        except (ValueError, ZeroDivisionError) as e:
            warnings.warn(f"AIPW computation failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan

    return bootstrap_estimates


def wboot_drdid_rc2(y, post, d, x, i_weights, n_bootstrap=1000, trim_level=0.995, random_state=None):
    r"""Compute bootstrap estimates for locally efficient doubly-robust DiD with repeated cross-sections.

    Implements the bootstrap inference for the locally efficient doubly-robust
    difference-in-differences estimator with repeated cross-section data. This version uses
    outcome regression on both treatment and control groups.

    Parameters
    ----------
    y : ndarray
        A 1D array representing the outcome variable for each unit.
    post : ndarray
        A 1D array representing the post-treatment period indicator (1 for post, 0 for pre)
        for each unit.
    d : ndarray
        A 1D array representing the treatment indicator (1 for treated, 0 for control)
        for each unit.
    x : ndarray
        A 2D array of covariates (including intercept if desired) with shape (n_units, n_features).
    i_weights : ndarray
        A 1D array of individual observation weights for each unit.
    n_bootstrap : int
        Number of bootstrap iterations. Default is 1000.
    trim_level : float
        Maximum propensity score value for control units to avoid extreme weights.
        Default is 0.995.
    random_state : int, RandomState instance or None
        Controls the random number generation for reproducibility.

    Returns
    -------
    ndarray
        A 1D array of bootstrap ATT estimates with length n_bootstrap.

    See Also
    --------
    aipw_did_rc_imp2 : The underlying AIPW estimator for repeated cross-sections.
    wboot_drdid_rc_imp2 : Improved bootstrap for doubly-robust DiD.
    """
    n_units = _validate_inputs({"y": y, "post": post, "d": d, "i_weights": i_weights}, x, n_bootstrap, trim_level)

    rng = np.random.RandomState(random_state)
    bootstrap_estimates = np.zeros(n_bootstrap)

    for b in range(n_bootstrap):
        v = rng.exponential(scale=1.0, size=n_units)
        b_weights = i_weights * v

        try:
            ps_b = _fit_logistic_boot_rc(d, x, b_weights)
        except (ValueError, np.linalg.LinAlgError) as e:
            warnings.warn(f"Propensity score estimation failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan
            continue

        ps_b = np.clip(ps_b, 1e-6, 1 - 1e-6)

        trim_ps = np.ones_like(ps_b, dtype=bool)
        control_mask = d == 0
        trim_ps[control_mask] = ps_b[control_mask] < trim_level

        try:
            control_pre_results = wols_rc(y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=True, treat=False)
            out_y_cont_pre_b = control_pre_results.out_reg

            control_post_results = wols_rc(
                y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=False, treat=False
            )
            out_y_cont_post_b = control_post_results.out_reg

            treat_pre_results = wols_rc(y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=True, treat=True)
            out_y_treat_pre_b = treat_pre_results.out_reg

            treat_post_results = wols_rc(y=y, post=post, d=d, x=x, ps=ps_b, i_weights=b_weights, pre=False, treat=True)
            out_y_treat_post_b = treat_post_results.out_reg

        except (ValueError, np.linalg.LinAlgError, KeyError) as e:
            warnings.warn(f"Outcome regression failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan
            continue

        b_weights_trimmed = b_weights.copy()
        b_weights_trimmed[~trim_ps] = 0

        try:
            att_b = aipw_did_rc_imp2(
                y=y,
                post=post,
                d=d,
                ps=ps_b,
                out_y_treat_post=out_y_treat_post_b,
                out_y_treat_pre=out_y_treat_pre_b,
                out_y_cont_post=out_y_cont_post_b,
                out_y_cont_pre=out_y_cont_pre_b,
                i_weights=b_weights_trimmed,
            )
            bootstrap_estimates[b] = att_b
        except (ValueError, ZeroDivisionError) as e:
            warnings.warn(f"AIPW computation failed in bootstrap {b}: {e}", UserWarning)
            bootstrap_estimates[b] = np.nan

    return bootstrap_estimates


def _fit_logistic_boot_rc(d, x, b_weights):
    """Dispatch logistic regression for bootstrap to GPU or statsmodels."""
    xp = get_backend()
    if xp is not np:
        _, ps = cupy_logistic_irls(
            xp.asarray(d, dtype=xp.float64),
            xp.asarray(x, dtype=xp.float64),
            xp.asarray(b_weights, dtype=xp.float64),
        )
        return to_numpy(ps)
    logit_model = sm.GLM(d, x, family=sm.families.Binomial(), freq_weights=b_weights)
    logit_results = logit_model.fit()
    return logit_results.predict(x)



"""Utility functions for preprocessing."""

import re
import warnings

import numpy as np
import polars as pl
import scipy.linalg

from ..dataframe import to_polars


def map_to_idx(vals, time_map):
    """Map values to indices."""
    vals_arr = np.asarray(vals, dtype=float)
    if vals_arr.ndim == 0:
        val_item = vals_arr.item()
        if np.isinf(val_item):
            return val_item
        return time_map.get(val_item, val_item)

    result = np.empty(len(vals_arr), dtype=float)
    for i, v in enumerate(vals_arr):
        if np.isinf(v):
            result[i] = v
        else:
            result[i] = time_map.get(v, v)

    if not np.any(np.isinf(result)):
        return result.astype(int)
    return result


def make_balanced_panel(data, idname, tname):
    """Make balanced panel.

    Parameters
    ----------
    data : pd.DataFrame or pl.DataFrame
        Input panel data.
    idname : str
        Name of the unit identifier column.
    tname : str
        Name of the time period column.

    Returns
    -------
    pl.DataFrame
        Balanced panel data containing only units observed in all time periods.
    """
    df = to_polars(data)
    if df.is_empty():
        return df

    n_periods = df[tname].n_unique()
    counts = df.group_by(idname).len()
    complete_ids = counts.filter(pl.col("len") == n_periods)[idname].to_list()
    return df.filter(pl.col(idname).is_in(complete_ids))


def get_first_difference(df, idname, yname, tname):
    """Get first difference.

    Parameters
    ----------
    df : pd.DataFrame or pl.DataFrame
        Input data.
    idname : str
        Name of unit identifier column.
    yname : str
        Name of outcome column.
    tname : str
        Name of time column.

    Returns
    -------
    pl.DataFrame
        DataFrame with original columns plus 'dy' column containing first differences.
    """
    data = to_polars(df)
    data = data.sort([idname, tname])
    return data.with_columns((pl.col(yname) - pl.col(yname).shift(1).over(idname)).alias("dy"))


def get_group(df, idname, tname, treatname, treat_period=None):
    """Get group.

    Parameters
    ----------
    df : pd.DataFrame or pl.DataFrame
        Input data.
    idname : str
        Name of unit identifier column.
    tname : str
        Name of time column.
    treatname : str
        Name of treatment column.
    treat_period : int or None
        Known treatment onset period. When provided, units with any positive
        value of *treatname* are assigned ``G = treat_period`` and all others
        receive ``G = 0``, bypassing the first-switch detection logic.

    Returns
    -------
    pl.DataFrame
        DataFrame with original columns plus 'G' column containing group assignment.
    """
    data = to_polars(df)

    if treat_period is not None:
        ever_treated = (
            data.group_by(idname)
            .agg((pl.col(treatname) > 0).any().alias("_ever"))
            .with_columns(
                pl.when(pl.col("_ever"))
                .then(pl.lit(treat_period, dtype=pl.Int64))
                .otherwise(pl.lit(0, dtype=pl.Int64))
                .alias("G")
            )
            .drop("_ever")
        )
        return data.join(ever_treated, on=idname, how="left")

    df_sorted = data.sort([idname, tname])

    df_with_treat = df_sorted.with_columns(
        (pl.col(treatname) > 0).alias("_is_treated"),
        (pl.col(treatname) > 0).cum_sum().over(idname).alias("_treat_cumsum"),
    ).with_columns(((pl.col("_treat_cumsum") == 1) & pl.col("_is_treated")).alias("_first_treat"))

    first_treat_df = (
        df_with_treat.filter(pl.col("_first_treat")).group_by(idname).agg(pl.col(tname).first().alias("_group"))
    )

    result = data.join(first_treat_df, on=idname, how="left").with_columns(
        pl.col("_group").fill_null(0).cast(pl.Int64).alias("G")
    )

    return result.drop("_group")


def two_by_two_subset(
    data,
    g,
    tp,
    control_group="notyettreated",
    anticipation=0,
    base_period="varying",
):
    """Two by two subset for treatment DiD.

    Parameters
    ----------
    data : pd.DataFrame or pl.DataFrame
        Input data with columns 'G', 'period', 'id'.
    g : numeric
        Treatment group.
    tp : numeric
        Time period.
    control_group : str
        Control group type ('notyettreated' or 'nevertreated').
    anticipation : int
        Anticipation periods.
    base_period : str
        Base period type ('varying' or 'universal').

    Returns
    -------
    dict
        Dictionary with 'gt_data', 'n1', 'disidx'.
    """
    df = to_polars(data)
    main_base_period = g - anticipation - 1

    if base_period == "varying":
        base_period_val = tp - 1 if tp < (g - anticipation) else main_base_period
    else:
        base_period_val = main_base_period

    if control_group == "notyettreated":
        unit_mask = (pl.col("G") == g) | (pl.col("G") > tp)
    else:
        unit_mask = (pl.col("G") == g) | pl.col("G").is_infinite()

    this_data = df.filter(unit_mask)

    time_mask = (pl.col("period") == tp) | (pl.col("period") == base_period_val)
    this_data = this_data.filter(time_mask)

    this_data = this_data.with_columns(
        pl.when(pl.col("period") == tp).then(pl.lit("post")).otherwise(pl.lit("pre")).alias("name"),
        (pl.col("G") == g).cast(pl.Int64).alias("D"),
    )

    if this_data["D"].n_unique() < 2:
        return {"gt_data": pl.DataFrame(), "n1": 0, "disidx": np.array([])}

    n1 = this_data["id"].n_unique()
    all_ids = df["id"].unique().to_numpy()
    subset_ids = this_data["id"].unique().to_numpy()
    disidx = np.isin(all_ids, subset_ids)

    return {"gt_data": this_data, "n1": n1, "disidx": disidx}


def choose_knots_quantile(x, num_knots):
    """Choose knots quantile."""
    if num_knots <= 0:
        return np.array([])

    x = np.asarray(x)
    if len(x) == 0:
        return np.array([])

    probs = np.linspace(0, 1, num_knots + 2)
    quantiles = np.quantile(x, probs)
    return quantiles[1:-1]


def create_dose_grid(dose_values, n_points=50):
    """Create dose grid."""
    dose_values = np.asarray(dose_values)
    positive_doses = dose_values[dose_values > 0]

    if len(positive_doses) == 0:
        return np.array([])

    return np.linspace(positive_doses.min(), positive_doses.max(), n_points)


def validate_dose_values(dose, treatment_group, never_treated_value=float("inf")):
    """Validate dose values."""
    dose = np.asarray(dose)
    treatment_group = np.asarray(treatment_group)

    errors = []
    warnings = []

    if (dose < 0).any():
        errors.append("Negative dose values detected")

    never_treated = treatment_group == never_treated_value
    never_treated_with_dose = never_treated & (dose > 0)
    if never_treated_with_dose.any():
        n_issues = never_treated_with_dose.sum()
        warnings.append(f"{n_issues} never-treated units have positive dose values")

    treated = (treatment_group != never_treated_value) & (treatment_group > 0)
    treated_no_dose = treated & (dose == 0)
    if treated_no_dose.any():
        n_issues = treated_no_dose.sum()
        warnings.append(f"{n_issues} treated units have zero dose values")

    return {
        "is_valid": len(errors) == 0,
        "errors": errors,
        "warnings": warnings,
    }


def parse_formula(formula):
    """Parse formula string to extract components."""
    parts = formula.split("~")
    if len(parts) != 2:
        raise ValueError("Formula must be in the form 'y ~ x1 + x2 + ...'")

    outcome = parts[0].strip()
    predictors_str = parts[1].strip()

    var_pattern = r"\b[a-zA-Z_]\w*\b"
    all_vars = re.findall(var_pattern, predictors_str)

    exclude = {"C", "I", "Q", "bs", "ns", "log", "exp", "sqrt", "abs", "np"}
    predictors = [v for v in all_vars if v not in exclude]

    seen = set()
    predictors = [x for x in predictors if not (x in seen or seen.add(x))]

    return {
        "outcome": outcome,
        "predictors": predictors,
        "formula": formula,
    }


def extract_vars_from_formula(formula):
    """Extract all variable names from formula string."""
    parsed = parse_formula(formula)
    vars_list = []
    if parsed["outcome"]:
        vars_list.append(parsed["outcome"])
    vars_list.extend(parsed["predictors"])
    return vars_list


def is_balanced_panel(data, tname, idname):
    """Check if the panel data is balanced.

    Parameters
    ----------
    data : pd.DataFrame or pl.DataFrame
        The input data.
    tname : str
        Name of time column.
    idname : str
        Name of id column.

    Returns
    -------
    bool
        True if panel is balanced (all units observed in all periods).
    """
    df = to_polars(data)
    n_periods = df[tname].n_unique()
    obs_per_unit = df.group_by(idname).agg(pl.col(tname).n_unique().alias("n_obs"))

    return (obs_per_unit["n_obs"] == n_periods).all()


def add_intercept(covariates):
    """Add intercept column to covariate matrix.

    Parameters
    ----------
    covariates : ndarray or None
        Covariate matrix.

    Returns
    -------
    ndarray or None
        Covariate matrix with intercept column prepended, or None if input is None.
    """
    if covariates is None or covariates.shape[1] == 0:
        return None

    intercept = np.ones((covariates.shape[0], 1))
    return np.hstack([intercept, covariates])


def extract_covariates(data, xformla):
    """Extract covariate matrix from DataFrame given a formula.

    Parameters
    ----------
    data : pd.DataFrame or pl.DataFrame
        The input data.
    xformla : str or None
        Formula for covariates in the form "~ x1 + x2 + x3".

    Returns
    -------
    ndarray or None
        Covariate matrix with intercept, or None if no covariates.
    """
    if xformla is None or xformla == "~1":
        return None

    df = to_polars(data)

    formula_str = xformla.strip()
    if formula_str.startswith("~"):
        formula_str = "y " + formula_str

    parsed = parse_formula(formula_str)
    covariate_names = parsed["predictors"]

    if not covariate_names or covariate_names == ["1"]:
        return None

    covariate_names = [c for c in covariate_names if c != "1"]

    missing_covs = [c for c in covariate_names if c not in df.columns]
    if missing_covs:
        raise ValueError(f"Covariates not found in data: {missing_covs}")

    X = df.select(covariate_names).to_numpy()
    intercept = np.ones((X.shape[0], 1))
    return np.hstack([intercept, X])


def get_covariate_names_from_formula(xformla):
    """Extract covariate names from an xformla string, returning None if no covariates."""
    if not xformla or xformla == "~1":
        return None
    return extract_vars_from_formula(xformla)


def remove_collinear(cov_matrix, var_names, tol=1e-6):
    """Remove collinear columns from covariate matrix using QR decomposition."""
    if cov_matrix.shape[1] == 0:
        return cov_matrix, var_names

    _, r, pivot = scipy.linalg.qr(cov_matrix, mode="economic", pivoting=True)

    diag_r = np.abs(np.diag(r))
    rank = np.sum(diag_r > tol * diag_r[0]) if len(diag_r) > 0 else 0

    keep_indices = pivot[:rank]
    keep_indices = np.sort(keep_indices)

    kept_vars = [var_names[i] for i in keep_indices]
    return cov_matrix[:, keep_indices], kept_vars


def check_partition_collinearity(cov_matrix, subgroup, var_names, tol=1e-6):
    """Check for partition-specific collinearity in DDD comparisons."""
    if len(var_names) == 0:
        return {}, []

    comparison_groups = [3, 2, 1]
    partition_collinear: dict[str, list[str]] = {}

    for comp_group in comparison_groups:
        mask = (subgroup == 4) | (subgroup == comp_group)
        cov_subset = cov_matrix[mask]

        if cov_subset.shape[0] == 0:
            continue

        _, kept_vars = remove_collinear(cov_subset, list(var_names), tol=tol)
        kept_set = set(kept_vars)
        collinear_in_subset = [v for v in var_names if v not in kept_set]

        if collinear_in_subset:
            partition_name = f"subgroup 4 vs {comp_group}"
            for var in collinear_in_subset:
                if var not in partition_collinear:
                    partition_collinear[var] = []
                partition_collinear[var].append(partition_name)

    return partition_collinear, list(partition_collinear.keys())


def create_ddd_subgroups(treat, partition, treat_val):
    """Create subgroup assignments for DDD.

    Subgroup definitions:
    - 4: Treated AND Eligible (treat=g, partition=1)
    - 3: Treated BUT Ineligible (treat=g, partition=0)
    - 2: Eligible BUT Untreated (treat=0, partition=1)
    - 1: Untreated AND Ineligible (treat=0, partition=0)
    """
    is_treated = treat == treat_val
    is_eligible = partition == 1

    subgroup = np.where(
        is_treated & is_eligible,
        4,
        np.where(is_treated & ~is_eligible, 3, np.where(is_eligible, 2, 1)),
    )
    return subgroup


def validate_subgroup_sizes(subgroup_counts, min_size=5):
    """Validate that each subgroup has sufficient observations."""
    for sg, count in subgroup_counts.items():
        if count < min_size:
            raise ValueError(f"Subgroup {sg} has only {count} observations. Minimum required is {min_size}.")


def extract_ddd_covariates(df: pl.DataFrame, xformla, subgroup: np.ndarray | None = None):
    """Extract and process covariates from data for DDD."""
    if xformla == "~1":
        n_units = len(df.filter(pl.col("_post") == 0))
        return np.empty((n_units, 0)), []

    covariate_vars = extract_vars_from_formula(xformla)

    df_pre = df.filter(pl.col("_post") == 0)
    cov_matrix = df_pre.select(covariate_vars).to_numpy().astype(float)

    if np.any(np.isnan(cov_matrix)):
        warnings.warn(
            "Missing values in covariates. Rows with NaN will cause issues.",
            stacklevel=3,
        )

    cov_matrix, kept_vars = remove_collinear(cov_matrix, covariate_vars)

    if subgroup is not None and len(kept_vars) > 0:
        partition_collinear, all_collinear = check_partition_collinearity(cov_matrix, subgroup, kept_vars)

        if all_collinear:
            partition_warnings = [
                f"  - {var} (collinear in: {', '.join(partitions)})" for var, partitions in partition_collinear.items()
            ]
            warnings.warn(
                "The following covariates were dropped due to partition-specific collinearity:\n"
                + "\n".join(partition_warnings),
                stacklevel=3,
            )

            keep_mask = [var not in all_collinear for var in kept_vars]
            cov_matrix = cov_matrix[:, keep_mask]
            kept_vars = [v for v in kept_vars if v not in all_collinear]

    return cov_matrix, kept_vars



"""Shared bootstrap functions for distributed backends."""

from __future__ import annotations

import warnings

import numpy as np


def _local_bootstrap(inf_func_local, biters, seed):
    """Compute local bootstrap contributions on one partition."""
    rng = np.random.default_rng(seed)
    n_local = inf_func_local.shape[0]
    k = inf_func_local.shape[1]

    p_kappa = 0.5 * (1 + np.sqrt(5)) / np.sqrt(5)
    k1 = 0.5 * (1 - np.sqrt(5))
    k2 = 0.5 * (1 + np.sqrt(5))

    local_bres = np.zeros((biters, k))
    for b in range(biters):
        draws = rng.binomial(1, p_kappa, size=n_local)
        v = np.where(draws == 1, k1, k2)
        local_bres[b] = np.sum(inf_func_local * v[:, None], axis=0)

    return local_bres, np.zeros(k), n_local


def _sum_bootstrap_pair(a, b):
    """Sum two (local_bres, zeros, n) tuples."""
    return a[0] + b[0], a[1] + b[1], a[2] + b[2]


def compute_bootstrap_se(total_bres, n_total, alpha=0.05):
    r"""Compute bootstrap standard errors and critical value from aggregated bootstrap draws.

    Parameters
    ----------
    total_bres : ndarray of shape (biters, k)
        Aggregated bootstrap draws (already summed across partitions).
    n_total : int
        Total number of observations.
    alpha : float, default 0.05
        Significance level for confidence intervals.

    Returns
    -------
    bres : ndarray of shape (biters, k)
        Scaled bootstrap results.
    se : ndarray of shape (k,)
        Standard errors.
    crit_val : float
        Critical value for uniform confidence bands.
    """
    bres = np.sqrt(n_total) * total_bres / n_total

    k = bres.shape[1]
    col_sums_sq = np.sum(bres**2, axis=0)
    ndg_dim = (~np.isnan(col_sums_sq)) & (col_sums_sq > np.sqrt(np.finfo(float).eps) * 10)
    bres_clean = bres[:, ndg_dim]

    se_full = np.full(k, np.nan)
    crit_val = np.nan

    if bres_clean.shape[1] > 0:
        q75 = np.percentile(bres_clean, 75, axis=0)
        q25 = np.percentile(bres_clean, 25, axis=0)
        b_sigma = (q75 - q25) / 1.3489795
        b_sigma[b_sigma <= np.sqrt(np.finfo(float).eps) * 10] = np.nan
        se_full[ndg_dim] = b_sigma / np.sqrt(n_total)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            b_t = np.max(np.abs(bres_clean / b_sigma), axis=1)

        b_t_finite = b_t[np.isfinite(b_t)]
        if len(b_t_finite) > 0:
            crit_val = np.percentile(b_t_finite, 100 * (1 - alpha))

    return bres, se_full, crit_val



"""Dask-distributed intertemporal DID estimation."""

from __future__ import annotations

import functools
from types import SimpleNamespace

import distributed
import numpy as np
import polars as pl
from scipy import stats

from moderndid.core.preprocess.utils import get_covariate_names_from_formula
from moderndid.dask._gram import tree_reduce
from moderndid.didinter.compute_did_multiplegt import (
    _compute_ate,
    _run_het_regression,
    _test_effects_equality,
)
from moderndid.didinter.results import (
    DIDInterResult,
    EffectsResult,
    PlacebosResult,
)
from moderndid.didinter.variance import compute_clustered_variance, compute_joint_test
from moderndid.distributed._didinter_partition import (
    apply_trends_lin_accumulation,
    build_didinter_partition_arrays,
    partition_apply_control_adjustment,
    partition_apply_globals,
    partition_check_group_level_covariates,
    partition_compute_variance_part2,
    partition_control_gram,
    partition_control_influence_sums,
    partition_delta_and_variance,
    partition_extract_het_data,
    partition_group_sums_and_scalars,
    partition_horizon_covariate_ops,
    partition_horizon_local_ops,
    partition_influence_and_meta,
    partition_variance_influence,
    prepare_het_sample,
    reduce_control_gram,
    reduce_control_influence_sums,
    reduce_dof_stats,
    reduce_global_scalars,
    reduce_group_sums,
    solve_control_coefficients,
)
from moderndid.distributed._didinter_preprocess import (
    cap_effects_placebo,
    partition_extract_metadata,
    partition_preprocess_global,
    partition_preprocess_local,
    reduce_metadata,
    validate_distributed,
)


def dask_did_multiplegt_mp(
    client,
    data,
    yname,
    tname,
    idname,
    dname,
    cluster,
    weightsname,
    xformla,
    effects,
    placebo,
    normalized,
    effects_equal,
    predict_het,
    switchers,
    only_never_switchers,
    same_switchers,
    same_switchers_pl,
    trends_lin,
    trends_nonparam,
    continuous,
    ci_level,
    less_conservative_se,
    keep_bidirectional_switchers,
    drop_missing_preswitch,
    boot,
    biters,
    random_state,
):
    """Run the multi-partition intertemporal DiD estimator on a Dask cluster.

    Preprocesses the panel data on Dask workers (no driver collection),
    then distributes horizon-level effect and placebo estimation across
    workers. Covariates, linear trends, nonparametric trends, and
    heterogeneity prediction are all handled distributedly.

    Parameters
    ----------
    client : distributed.Client
        Active Dask distributed client.
    data : dask.dataframe.DataFrame
        Panel data in long format as a Dask DataFrame.
    yname : str
        Name of the outcome variable.
    tname : str
        Name of the time period variable.
    idname : str
        Name of the unit identifier variable.
    dname : str
        Name of the treatment variable.
    cluster : str or None
        Cluster variable for standard errors.
    weightsname : str or None
        Name of the sampling weights variable.
    xformla : str
        Formula for covariates (e.g., ``"~X1+X2"``).
    effects : int
        Number of post-treatment horizons to estimate.
    placebo : int
        Number of pre-treatment placebo horizons.
    normalized : bool
        If True, normalize estimates by cumulative treatment change.
    effects_equal : bool
        If True, test equality of effects across horizons.
    predict_het : str or None
        Variable for heterogeneity prediction.
    switchers : str
        Switcher direction filter (``""``, ``"up"``, or ``"down"``).
    only_never_switchers : bool
        If True, use only never-switchers as the control group.
    same_switchers : bool
        If True, keep the same switcher composition across effect horizons.
    same_switchers_pl : bool
        If True, keep the same switcher composition across placebo horizons.
    trends_lin : bool
        If True, include unit-specific linear trends.
    trends_nonparam : str or None
        Variable for non-parametric trends.
    continuous : int
        Continuous treatment flag.
    ci_level : float
        Confidence interval level in percent (e.g., 95.0).
    less_conservative_se : bool
        If True, use less conservative standard error estimator.
    keep_bidirectional_switchers : bool
        If True, retain units that switch treatment in both directions.
    drop_missing_preswitch : bool
        If True, drop observations with missing pre-switch data.
    boot : bool
        Whether to use bootstrap inference.
    biters : int
        Number of bootstrap iterations.
    random_state : int or None
        Random seed for reproducibility.

    Returns
    -------
    DIDInterResult
        Estimation results containing effects, placebos, ATE, influence
        functions, and diagnostic statistics.
    """
    _ = (boot, biters, random_state)
    covariate_names = get_covariate_names_from_formula(xformla)
    trend_vars = list(trends_nonparam) if trends_nonparam else None

    col_config = {
        "gname": idname,
        "tname": tname,
        "yname": yname,
        "dname": dname,
        "cluster": cluster,
        "covariate_names": covariate_names if covariate_names else None,
        "trends_nonparam": trend_vars,
        "het_covariates": predict_het[0] if predict_het else None,
    }

    config_flags = {
        "gname": idname,
        "tname": tname,
        "yname": yname,
        "dname": dname,
        "weightsname": weightsname,
        "xformla": xformla,
        "trends_nonparam": trends_nonparam,
        "switchers": switchers,
        "keep_bidirectional_switchers": keep_bidirectional_switchers,
        "drop_missing_preswitch": drop_missing_preswitch,
        "continuous": continuous,
        "trends_lin": trends_lin,
        "same_switchers": same_switchers,
        "same_switchers_pl": same_switchers_pl,
        "effects": effects,
        "placebo": placebo,
        "only_never_switchers": only_never_switchers,
        "normalized": normalized,
        "less_conservative_se": less_conservative_se,
    }

    validate_distributed(data.columns.tolist(), col_config, config_flags)

    n_workers = max(2, len(client.scheduler_info()["workers"]))

    het_covariates = predict_het[0] if predict_het else None

    needed_cols = [yname, tname, idname, dname]
    if weightsname:
        needed_cols.append(weightsname)
    if cluster:
        needed_cols.append(cluster)
    if covariate_names:
        needed_cols.extend(covariate_names)
    if trends_nonparam:
        needed_cols.extend(trends_nonparam)
    if het_covariates:
        needed_cols.extend(het_covariates)
    needed_cols = list(dict.fromkeys(c for c in needed_cols if c in data.columns))
    ddf = data[needed_cols]

    try:
        ddf = ddf.shuffle(idname, npartitions=n_workers)
    except (AttributeError, TypeError):
        ddf = ddf.set_index(idname, sorted=False, npartitions=n_workers).reset_index()

    delayed_parts = ddf.to_delayed()
    pdf_futures = client.compute(delayed_parts)

    phase_a_futures = [client.submit(_preprocess_local_from_pandas, pf, col_config, config_flags) for pf in pdf_futures]
    distributed.wait(phase_a_futures)

    meta_futures = [client.submit(partition_extract_metadata, pf, config_flags) for pf in phase_a_futures]
    meta_list = client.gather(meta_futures)
    global_metadata = functools.reduce(reduce_metadata, meta_list)

    effects, placebo = cap_effects_placebo(config_flags, global_metadata)

    n_groups = len(global_metadata["all_gnames_first_obs"])
    t_max = int(global_metadata["t_max"])
    n_switchers_total = global_metadata["n_switchers"]
    n_never_switchers_total = global_metadata["n_never_switchers"]

    part_futures = [
        client.submit(_apply_global_preprocess, pf, global_metadata, col_config, config_flags) for pf in phase_a_futures
    ]

    nrow_futures = [client.submit(_get_nrows, pf) for pf in part_futures]
    nrow_results = client.gather(nrow_futures)
    part_futures = [pf for pf, nr in zip(part_futures, nrow_results, strict=False) if nr > 0]

    config_dict = {
        "switchers": switchers,
        "only_never_switchers": only_never_switchers,
        "same_switchers": same_switchers,
        "normalized": normalized,
        "less_conservative_se": less_conservative_se,
    }

    sorted_gnames = np.array(sorted(global_metadata["all_gnames_first_obs"]))
    gname_to_idx = {g: i for i, g in enumerate(sorted_gnames)}

    cluster_id_map = None
    if cluster:
        cid_futures = [client.submit(_extract_cluster_ids, pf) for pf in part_futures]
        cid_maps = client.gather(cid_futures)
        cluster_id_map = {}
        for cm in cid_maps:
            cluster_id_map.update(cm)

    effects_results = _distributed_did_effects(
        client,
        part_futures,
        config_dict,
        sorted_gnames,
        gname_to_idx,
        n_horizons=effects,
        n_groups=n_groups,
        t_max=t_max,
        horizon_type="effect",
        normalized=normalized,
        cluster_col=cluster,
        less_conservative_se=less_conservative_se,
        covariate_names=covariate_names if covariate_names else None,
        trend_vars=trend_vars,
        trends_lin=trends_lin,
        cluster_id_map=cluster_id_map,
    )

    placebos_results = None
    if placebo > 0:
        placebos_results = _distributed_did_effects(
            client,
            part_futures,
            config_dict,
            sorted_gnames,
            gname_to_idx,
            n_horizons=placebo,
            n_groups=n_groups,
            t_max=t_max,
            horizon_type="placebo",
            normalized=normalized,
            cluster_col=cluster,
            less_conservative_se=less_conservative_se,
            covariate_names=covariate_names if covariate_names else None,
            trend_vars=trend_vars,
            trends_lin=trends_lin,
            cluster_id_map=cluster_id_map,
        )

    heterogeneity = None
    if predict_het and not normalized and effects_results:
        heterogeneity = _distributed_heterogeneity(
            client,
            part_futures,
            effects,
            predict_het,
            trends_nonparam,
            trends_lin,
        )

    alpha = 1 - ci_level / 100
    z_crit = stats.norm.ppf(1 - alpha / 2)

    ate = _compute_ate(effects_results, z_crit, n_groups) if effects_results else None

    effects_equal_test = None
    if effects_equal and effects > 1 and effects_results:
        effects_equal_test = _test_effects_equality(effects_results)

    placebo_joint_test = None
    if placebo > 1 and placebos_results is not None:
        placebo_joint_test = compute_joint_test(
            placebos_results["estimates"],
            placebos_results["vcov"],
        )

    effects_obj = EffectsResult(
        horizons=effects_results["horizons"],
        estimates=effects_results["estimates"],
        std_errors=effects_results["std_errors"],
        ci_lower=effects_results["estimates"] - z_crit * effects_results["std_errors"],
        ci_upper=effects_results["estimates"] + z_crit * effects_results["std_errors"],
        n_switchers=effects_results["n_switchers"],
        n_observations=effects_results["n_observations"],
    )

    placebos_obj = None
    if placebos_results is not None:
        placebos_obj = PlacebosResult(
            horizons=placebos_results["horizons"],
            estimates=placebos_results["estimates"],
            std_errors=placebos_results["std_errors"],
            ci_lower=placebos_results["estimates"] - z_crit * placebos_results["std_errors"],
            ci_upper=placebos_results["estimates"] + z_crit * placebos_results["std_errors"],
            n_switchers=placebos_results["n_switchers"],
            n_observations=placebos_results["n_observations"],
        )

    return DIDInterResult(
        effects=effects_obj,
        placebos=placebos_obj,
        ate=ate,
        n_units=n_switchers_total + n_never_switchers_total,
        n_switchers=n_switchers_total,
        n_never_switchers=n_never_switchers_total,
        ci_level=ci_level,
        effects_equal_test=effects_equal_test,
        placebo_joint_test=placebo_joint_test,
        influence_effects=effects_results.get("influence_func"),
        influence_placebos=placebos_results.get("influence_func") if placebos_results else None,
        heterogeneity=heterogeneity,
        estimation_params={
            "effects": effects,
            "placebo": placebo,
            "normalized": normalized,
            "switchers": switchers,
            "xformla": xformla,
            "cluster": cluster,
            "trends_lin": trends_lin,
            "trends_nonparam": trends_nonparam,
            "only_never_switchers": only_never_switchers,
            "same_switchers": same_switchers,
            "same_switchers_pl": same_switchers_pl,
            "continuous": continuous,
            "weightsname": weightsname,
        },
    )


def _distributed_did_effects(
    client,
    part_futures,
    config_dict,
    sorted_gnames,
    gname_to_idx,
    n_horizons,
    n_groups,
    t_max,
    horizon_type,
    normalized,
    cluster_col,
    less_conservative_se,
    covariate_names=None,
    trend_vars=None,
    trends_lin=False,
    cluster_id_map=None,
):
    """Estimate horizon-level treatment effects across Dask partitions.

    Iterates over each horizon, distributes local computations (group sums,
    influence functions, variance estimation) to workers via futures, and
    aggregates results on the driver using tree-reduce patterns. Supports
    covariates, nonparametric trends, and linear trends.

    Parameters
    ----------
    client : distributed.Client
        Active Dask distributed client.
    part_futures : list of distributed.Future
        Futures referencing partition arrays on workers.
    config_dict : dict
        Subset of config fields needed by partition functions.
    sorted_gnames : numpy.ndarray
        Sorted array of all unique unit identifiers.
    gname_to_idx : dict
        Mapping from unit identifier to index in ``sorted_gnames``.
    n_horizons : int
        Number of horizons (effects or placebos) to estimate.
    n_groups : int
        Total number of unique panel units.
    t_max : int
        Maximum time period in the data.
    horizon_type : {"effect", "placebo"}
        Whether to estimate treatment effects or placebo tests.
    normalized : bool
        Whether to normalize estimates.
    cluster_col : str or None
        Cluster column name.
    less_conservative_se : bool
        Whether to use less conservative SE.
    covariate_names : list of str or None
        Covariate names for control regression.
    trend_vars : list of str or None
        Non-parametric trend variable names for extended grouping.
    trends_lin : bool
        Whether to apply linear trend accumulation after all horizons.
    cluster_id_map : dict or None
        Pre-extracted ``{gname: cluster_id}`` mapping. Computed once
        by the caller to avoid pulling full partitions per horizon.

    Returns
    -------
    dict
        Dictionary with keys ``"horizons"``, ``"estimates"``,
        ``"estimates_unnorm"``, ``"std_errors"``, ``"n_switchers"``,
        ``"n_switchers_weighted"``, ``"delta_d_arr"``, ``"n_observations"``,
        ``"influence_func"``, ``"influence_func_unnorm"``, and ``"vcov"``.
    """
    horizons = np.arange(1, n_horizons + 1) if horizon_type == "effect" else -np.arange(1, n_horizons + 1)

    estimates = np.zeros(n_horizons)
    estimates_unnorm = np.zeros(n_horizons)
    std_errors = np.zeros(n_horizons)
    n_switchers_arr = np.zeros(n_horizons)
    n_switchers_weighted_arr = np.zeros(n_horizons)
    delta_d_arr = np.zeros(n_horizons)
    n_obs_arr = np.zeros(n_horizons)
    influence_funcs = []
    influence_funcs_unnorm = []

    n_controls = len(covariate_names) if covariate_names else 0

    for idx, h in enumerate(horizons):
        abs_h = abs(h)

        part_futures = [
            client.submit(partition_horizon_local_ops, pf, abs_h, horizon_type, config_dict, t_max)
            for pf in part_futures
        ]

        if covariate_names:
            part_futures = [
                client.submit(partition_horizon_covariate_ops, pf, abs_h, covariate_names) for pf in part_futures
            ]

            gram_futures = [client.submit(partition_control_gram, pf, abs_h, covariate_names) for pf in part_futures]
            global_gram = tree_reduce(client, gram_futures, reduce_control_gram)

            coefficients = solve_control_coefficients(global_gram, n_controls, n_groups)

            if coefficients:
                part_futures = [
                    client.submit(partition_apply_control_adjustment, pf, abs_h, covariate_names, coefficients)
                    for pf in part_futures
                ]
            else:
                coefficients = {}
        else:
            coefficients = {}

        combined_gs_sc_futures = [
            client.submit(partition_group_sums_and_scalars, pf, abs_h, trend_vars) for pf in part_futures
        ]
        combined_gs_sc = client.gather(combined_gs_sc_futures)
        gs_list = [c[0] for c in combined_gs_sc]
        sc_list = [c[1] for c in combined_gs_sc]
        global_group_sums = functools.reduce(reduce_group_sums, gs_list) if gs_list else {}
        global_scalars = (
            functools.reduce(reduce_global_scalars, sc_list)
            if sc_list
            else {"n_switchers_weighted": 0, "switcher_gnames": set()}
        )

        n_switchers_weighted = global_scalars["n_switchers_weighted"]
        n_switchers_unweighted = len(global_scalars["switcher_gnames"])

        if n_switchers_unweighted == 0:
            estimates[idx] = np.nan
            std_errors[idx] = np.nan
            n_switchers_arr[idx] = 0
            n_obs_arr[idx] = 0
            continue

        part_futures = [
            client.submit(partition_apply_globals, pf, abs_h, global_group_sums, trend_vars) for pf in part_futures
        ]

        combined_inf_meta_futures = [
            client.submit(
                partition_influence_and_meta, pf, abs_h, n_groups, n_switchers_weighted, cluster_col, trend_vars
            )
            for pf in part_futures
        ]
        combined_inf_meta = client.gather(combined_inf_meta_futures)

        inf_func_full = np.zeros(len(sorted_gnames))
        total_if_sum = 0.0
        switcher_gnames_with_weight = {}
        total_obs = 0
        dof_list = []
        for inf_result, sw_map, obs_count, dof in combined_inf_meta:
            gname_if, partial_sum = inf_result
            total_if_sum += partial_sum
            for gn, val in gname_if.items():
                if gn in gname_to_idx:
                    inf_func_full[gname_to_idx[gn]] = val
            switcher_gnames_with_weight.update(sw_map)
            total_obs += obs_count
            dof_list.append(dof)
        global_dof = functools.reduce(reduce_dof_stats, dof_list) if dof_list else {}

        did_estimate = total_if_sum / n_groups

        estimates_unnorm[idx] = did_estimate
        n_switchers_weighted_arr[idx] = max(n_switchers_weighted, 1e-10)

        influence_funcs_unnorm.append(inf_func_full.copy())

        delta_d = None
        if normalized or horizon_type == "effect":
            combined_dd_var_futures = [
                client.submit(
                    partition_delta_and_variance,
                    pf,
                    abs_h,
                    horizon_type,
                    switcher_gnames_with_weight,
                    n_groups,
                    n_switchers_weighted,
                    global_dof,
                    cluster_col,
                    less_conservative_se,
                    trend_vars,
                )
                for pf in part_futures
            ]
            combined_dd_var = client.gather(combined_dd_var_futures)

            total_dd = 0.0
            total_dd_weight = 0.0
            inf_var_full = np.zeros(len(sorted_gnames))
            for dd_result, var_result in combined_dd_var:
                contrib, weight = dd_result
                total_dd += contrib
                total_dd_weight += weight
                for gn, val in var_result.items():
                    if gn in gname_to_idx:
                        inf_var_full[gname_to_idx[gn]] = val
            delta_d = total_dd / total_dd_weight if total_dd_weight > 0 else None
        else:
            var_if_futures = [
                client.submit(
                    partition_variance_influence,
                    pf,
                    abs_h,
                    n_groups,
                    n_switchers_weighted,
                    global_dof,
                    cluster_col,
                    less_conservative_se,
                    trend_vars,
                )
                for pf in part_futures
            ]
            inf_var_full = np.zeros(len(sorted_gnames))
            for fut in var_if_futures:
                gname_var_if = fut.result()
                for gn, val in gname_var_if.items():
                    if gn in gname_to_idx:
                        inf_var_full[gname_to_idx[gn]] = val

        if horizon_type == "effect":
            delta_d_arr[idx] = delta_d if delta_d is not None else 0.0

        if normalized and delta_d is not None and delta_d != 0:
            did_estimate = did_estimate / delta_d

        estimates[idx] = did_estimate

        if covariate_names and coefficients:
            useful_coefficients = {d: c for d, c in coefficients.items() if c.get("useful", False)}
            if useful_coefficients:
                cis_futures = [
                    client.submit(
                        partition_control_influence_sums,
                        pf,
                        abs_h,
                        covariate_names,
                        coefficients,
                        n_groups,
                        n_switchers_weighted,
                        trend_vars,
                    )
                    for pf in part_futures
                ]
                global_cis = tree_reduce(client, cis_futures, reduce_control_influence_sums)

                global_M_total = global_cis.get("M_total", {})

                part2_futures = [
                    client.submit(
                        partition_compute_variance_part2,
                        pf,
                        abs_h,
                        covariate_names,
                        coefficients,
                        global_M_total,
                        n_groups,
                    )
                    for pf in part_futures
                ]

                for fut in part2_futures:
                    gname_part2 = fut.result()
                    for gn, val in gname_part2.items():
                        if gn in gname_to_idx:
                            inf_var_full[gname_to_idx[gn]] -= val

        if cluster_col and cluster_id_map:
            cluster_ids = np.zeros(len(sorted_gnames), dtype=object)
            for gn, cid in cluster_id_map.items():
                if gn in gname_to_idx:
                    cluster_ids[gname_to_idx[gn]] = cid
            std_error = compute_clustered_variance(inf_var_full, cluster_ids, n_groups)
        else:
            std_error = np.sqrt(np.sum(inf_var_full**2)) / n_groups

        inf_func = inf_var_full.copy()

        if normalized and delta_d is not None and delta_d != 0:
            std_error = std_error / delta_d
            inf_func = inf_func / delta_d

        influence_funcs.append(inf_func)
        std_errors[idx] = std_error
        n_switchers_arr[idx] = n_switchers_unweighted
        n_obs_arr[idx] = total_obs

    if trends_lin and len(influence_funcs) == n_horizons and n_horizons > 0:
        cluster_ids_for_trends = None
        if cluster_col and cluster_id_map:
            cluster_ids_for_trends = np.zeros(len(sorted_gnames), dtype=object)
            for gn, cid in cluster_id_map.items():
                if gn in gname_to_idx:
                    cluster_ids_for_trends[gname_to_idx[gn]] = cid

        estimates, std_errors, influence_funcs = apply_trends_lin_accumulation(
            estimates,
            std_errors,
            influence_funcs,
            n_groups,
            cluster_col=cluster_col,
            cluster_ids=cluster_ids_for_trends,
        )

    vcov = None
    if len(influence_funcs) == n_horizons and all(len(f) > 0 for f in influence_funcs):
        influence_matrix = np.column_stack(influence_funcs)
        vcov = np.cov(influence_matrix, rowvar=False) / n_groups

    return {
        "horizons": horizons.astype(float),
        "estimates": estimates,
        "estimates_unnorm": estimates_unnorm,
        "std_errors": std_errors,
        "n_switchers": n_switchers_arr,
        "n_switchers_weighted": n_switchers_weighted_arr,
        "delta_d_arr": delta_d_arr,
        "n_observations": n_obs_arr,
        "influence_func": np.column_stack(influence_funcs) if influence_funcs else None,
        "influence_func_unnorm": np.column_stack(influence_funcs_unnorm) if influence_funcs_unnorm else None,
        "vcov": vcov,
    }


def _distributed_heterogeneity(client, part_futures, effects, predict_het, trends_nonparam, trends_lin):
    """Collect group-level summaries and run WLS heterogeneity regressions."""
    covariates, het_effects = predict_het

    if not isinstance(covariates, list) or not isinstance(het_effects, list) or len(covariates) == 0:
        return None

    varies_futures = [client.submit(partition_check_group_level_covariates, pf, covariates) for pf in part_futures]
    all_varies = set()
    for fut in varies_futures:
        all_varies |= fut.result()

    valid_covariates = [c for c in covariates if c not in all_varies]
    if len(valid_covariates) == 0:
        return None

    het_futures = [
        client.submit(
            partition_extract_het_data,
            pf,
            effects,
            valid_covariates,
            trends_nonparam,
            trends_lin,
        )
        for pf in part_futures
    ]

    all_rows = []
    for fut in het_futures:
        row_list = fut.result()
        all_rows.extend(row_list)

    if len(all_rows) == 0:
        return None

    het_df = pl.DataFrame(all_rows)

    all_horizons = (
        list(range(1, effects + 1)) if -1 in het_effects else [h_ for h_ in het_effects if 1 <= h_ <= effects]
    )

    if len(all_horizons) == 0:
        return None

    config = SimpleNamespace(trends_nonparam=trends_nonparam)

    results = []
    for horizon in all_horizons:
        het_sample = prepare_het_sample(het_df, horizon, trends_lin)
        if het_sample is None or len(het_sample) < len(valid_covariates) + 5:
            continue

        het_result = _run_het_regression(het_sample, valid_covariates, horizon, config)
        if het_result is not None:
            results.append(het_result)

    return results if len(results) > 0 else None


def _preprocess_local_from_pandas(pdf, col_config, config_flags):
    """Convert pandas partition to polars and run partition-local preprocessing."""
    return partition_preprocess_local(pl.from_pandas(pdf), col_config, config_flags)


def _apply_global_preprocess(pdf, metadata, col_config, config_flags):
    """Apply global preprocessing with broadcast metadata and convert to NumPy partition dict."""
    if pdf is None or len(pdf) == 0:
        return None
    preprocessed = partition_preprocess_global(pdf, metadata, col_config, config_flags)
    if len(preprocessed) == 0:
        return None
    return build_didinter_partition_arrays(preprocessed, col_config)


def _get_nrows(part):
    """Return partition row count without transferring the full partition."""
    if part is None:
        return 0
    return part.get("n_rows", 0)


def _extract_cluster_ids(part):
    """Extract unit-to-cluster-ID mapping from a partition."""
    first_mask = part["first_obs_by_gp"] == 1.0
    if "cluster" in part:
        gn = part["gname"][first_mask]
        cl = part["cluster"][first_mask]
        return dict(zip(gn.tolist(), cl.tolist(), strict=False))
    return {}
